#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import infer_dtype
from .core import TensorBinOp, TensorConstant


class TensorCopysign(operands.Copysign, TensorBinOp):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorCopysign, self).__init__(_casting=casting, _err=err,
                                             _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        return False

    @classmethod
    def constant_cls(cls):
        return TensorCopysignConstant


class TensorCopysignConstant(operands.CopysignConstant, TensorConstant):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorCopysignConstant, self).__init__(_casting=casting, _err=err,
                                                     _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        return False


@infer_dtype(np.copysign)
def copysign(x1, x2, out=None, where=None, **kwargs):
    """
    Change the sign of x1 to that of x2, element-wise.

    If both arguments are arrays or sequences, they have to be of the same
    length. If `x2` is a scalar, its sign will be copied to all elements of
    `x1`.

    Parameters
    ----------
    x1 : array_like
        Values to change the sign of.
    x2 : array_like
        The sign of `x2` is copied to `x1`.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    out : array_like
        The values of `x1` with the sign of `x2`.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.copysign(1.3, -1).execute()
    -1.3
    >>> (1/mt.copysign(0, 1)).execute()
    inf
    >>> (1/mt.copysign(0, -1)).execute()
    -inf

    >>> mt.copysign([-1, 0, 1], -1.1).execute()
    array([-1., -0., -1.])
    >>> mt.copysign([-1, 0, 1], mt.arange(3)-1).execute()
    array([-1.,  0.,  1.])
    """
    op = TensorCopysign(**kwargs)
    return op(x1, x2, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .ne import NeOptimizer
from .cp import CpOptimizer
from ...execution.core import NUMEXPR_INSTALLED


class Optimizer(object):
    engine_dic = {'numexpr': NeOptimizer,
                  'cupy': CpOptimizer}

    def __init__(self, graph, engine=None):
        self._graph = graph
        if not engine:
            self._engine = 'numexpr' if NUMEXPR_INSTALLED else 'numpy'
        else:
            self._engine = engine

    def optimize(self, keys=None):
        self._graph.decompose()
        if self._engine == 'numpy':
            return
        optimizer = self.engine_dic[self._engine](self._graph)
        optimizer.optimize(keys=keys)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorArgReduction


class TensorArgmaxChunk(operands.ArgmaxChunk, TensorArgReduction):
    def __init__(self, axis=None, dtype=np.dtype(int), keepdims=None,
                 combine_size=None, offset=None, total_shape=None, **kw):
        super(TensorArgmaxChunk, self).__init__(_axis=axis, _dtype=dtype,
                                                _keepdims=keepdims, _combine_size=combine_size,
                                                _offset=offset, _total_shape=total_shape, **kw)


class TensorArgmaxCombine(operands.ArgmaxCombine, TensorArgReduction):
    def __init__(self, axis=None, dtype=np.dtype(int), keepdims=None, combine_size=None, **kw):
        super(TensorArgmaxCombine, self).__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims,
                                                  _combine_size=combine_size, **kw)


class TensorArgmax(operands.Argmax, TensorArgReduction):
    def __init__(self, axis=None, dtype=np.dtype(int), keepdims=None, combine_size=None, **kw):
        super(TensorArgmax, self).__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims,
                                           _combine_size=combine_size, **kw)

    @staticmethod
    def _get_op_types():
        return TensorArgmaxChunk, TensorArgmax, TensorArgmaxCombine


def argmax(a, axis=None, out=None, keepdims=None, combine_size=None):
    """
    Returns the indices of the maximum values along an axis.

    Parameters
    ----------
    a : array_like
        Input tensor.
    axis : int, optional
        By default, the index is into the flattened tensor, otherwise
        along the specified axis.
    out : Tensor, optional
        If provided, the result will be inserted into this tensor. It should
        be of the appropriate shape and dtype.
    combine_size: int, optional
        The number of chunks to combine.

    Returns
    -------
    index_array : Tensor of ints
        Tensor of indices into the tensor. It has the same shape as `a.shape`
        with the dimension along `axis` removed.

    See Also
    --------
    Tensor.argmax, argmin
    amax : The maximum value along a given axis.
    unravel_index : Convert a flat index into an index tuple.

    Notes
    -----
    In case of multiple occurrences of the maximum values, the indices
    corresponding to the first occurrence are returned.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> from mars.session import new_session

    >>> sess = new_session().as_default()

    >>> a = mt.arange(6).reshape(2,3)
    >>> a.execute()
    array([[0, 1, 2],
           [3, 4, 5]])
    >>> mt.argmax(a).execute()
    5
    >>> mt.argmax(a, axis=0).execute()
    array([1, 1, 1])
    >>> mt.argmax(a, axis=1).execute()
    array([2, 2])

    Indexes of the maximal elements of a N-dimensional tensor:

    >>> ind = mt.unravel_index(mt.argmax(a, axis=None), a.shape)
    >>> sess.run(ind)
    (1, 2)
    >>> a[ind].execute()  # TODO(jisheng): accomplish when fancy index on tensor is supported

    >>> b = mt.arange(6)
    >>> b[1] = 5
    >>> b.execute()
    array([0, 5, 2, 3, 4, 5])
    >>> mt.argmax(b).execute()  # Only the first occurrence is returned.
    1

    """
    op = TensorArgmax(axis=axis, dtype=np.dtype(int), keepdims=keepdims, combine_size=combine_size)
    return op(a, out=out)



# Generated automatically.  DO NOT EDIT!
NULL = 0
SCALAR = 1
TENSOR_DATA_SOURCE = 2
TENSOR_ONES = 3
TENSOR_ONES_LIKE = 4
TENSOR_ZEROS = 5
TENSOR_ZEROS_LIKE = 6
TENSOR_EMPTY = 7
TENSOR_EMPTY_LIKE = 8
TENSOR_FULL = 9
TENSOR_ARANGE = 10
TENSOR_INDICES = 11
TENSOR_DIAG = 12
TENSOR_EYE = 13
TENSOR_LINSPACE = 14
TENSOR_TRIU = 15
TENSOR_TRIL = 16
RAND_RAND = 41
RAND_RANDN = 42
RAND_RANDINT = 43
RAND_RANDOM_INTEGERS = 44
RAND_RANDOM_SAMPLE = 45
RAND_RANDOM = 46
RAND_RANF = 47
RAND_SAMPLE = 48
RAND_BYTES = 49
RAND_BETA = 50
RAND_BINOMIAL = 51
RAND_CHISQUARE = 52
RAND_CHOICE = 53
RAND_DIRICHLET = 54
RAND_EXPONENTIAL = 55
RAND_F = 56
RAND_GAMMA = 57
RAND_GEOMETRIC = 58
RAND_GUMBEL = 59
RAND_HYPERGEOMETRIC = 60
RAND_LAPLACE = 61
RAND_LOGISTIC = 62
RAND_LOGNORMAL = 63
RAND_LOGSERIES = 64
RAND_MULTINOMIAL = 65
RAND_MULTIVARIATE_NORMAL = 66
RAND_NEGATIVE_BINOMIAL = 67
RAND_NONCENTRAL_CHISQURE = 68
RAND_NONCENTRAL_F = 69
RAND_NORMAL = 70
RAND_PARETO = 71
RAND_PERMUTATION = 72
RAND_POSSION = 73
RAND_POWER = 74
RAND_RAYLEIGH = 75
RAND_SHUFFLE = 76
RAND_STANDARD_CAUCHY = 77
RAND_STANDARD_EXPONENTIAL = 78
RAND_STANDARD_GAMMMA = 79
RAND_STANDARD_NORMAL = 80
RAND_STANDARD_T = 81
RAND_TOMAXINT = 82
RAND_TRIANGULAR = 83
RAND_UNIFORM = 84
RAND_VONMISES = 85
RAND_WALD = 86
RAND_WEIBULL = 87
RAND_ZIPF = 88
ADD = 101
ADD_CONSTANT = 102
SUB = 103
SUB_CONSTANT = 104
MUL = 105
MUL_CONSTANT = 106
DIV = 107
DIV_CONSTANT = 108
TRUEDIV = 109
TDIV_CONSTANT = 110
FLOORDIV = 111
FDIV_CONSTANT = 112
POW = 113
POW_CONSTANT = 114
MOD = 115
MOD_CONSTANT = 116
FMOD = 117
FMOD_CONSTANT = 118
LOGADDEXP = 119
LAE_CONSTANT = 120
LOGADDEXP2 = 121
LAE2_CONSTANT = 122
NEGATIVE = 123
POSITIVE = 124
ABSOLUTE = 125
FABS = 126
ABS = 127
RINT = 128
SIGN = 129
CONJ = 130
EXP = 131
EXP2 = 132
LOG = 133
LOG2 = 134
LOG10 = 135
EXPM1 = 136
LOG1P = 137
SQRT = 138
SQUARE = 139
CBRT = 140
RECIPROCAL = 141
EQ = 142
EQ_CONSTANT = 143
NE = 144
NE_CONSTANT = 145
LT = 146
LT_CONSTANT = 147
LE = 148
LE_CONSTANT = 149
GT = 150
GT_CONSTANT = 151
GE = 152
GE_CONSTANT = 153
SIN = 154
COS = 155
TAN = 156
ARCSIN = 157
ARCCOS = 158
ARCTAN = 159
ARCTAN2 = 160
ARCT2_CONSTANT = 161
HYPOT = 162
HYPOT_CONSTANT = 163
SINH = 164
COSH = 165
TANH = 166
ARCSINH = 167
ARCCOSH = 168
ARCTANH = 169
DEG2RAD = 170
RAD2DEG = 171
BITAND = 172
BITAND_CONSTANT = 173
BITOR = 174
BITOR_CONSTANT = 175
BITXOR = 176
BITXOR_CONSTANT = 177
INVERT = 178
LSHIFT = 179
LSHIFT_CONSTANT = 180
RSHIFT = 181
RSHIFT_CONSTANT = 182
AND = 183
AND_CONSTANT = 184
OR = 185
OR_CONSTANT = 186
XOR = 187
XOR_CONSTANT = 188
NOT = 189
MAXIMUM = 190
MAXIMUM_CONSTANT = 191
MINIMUM = 192
MINIMUM_CONSTANT = 193
AROUND = 194
FLOAT_POWER = 195
FLOAT_POWER_CONSTANT = 196
FMAX = 197
FMAX_CONSTANT = 198
FMIN = 199
FMIN_CONSTANT = 200
ISFINITE = 201
ISINF = 202
ISNAN = 203
SIGNBIT = 204
COPYSIGN = 205
COPYSIGN_CONSTANT = 206
NEXTAFTER = 207
NEXTAFTER_CONSTANT = 208
SPACING = 209
LDEXP = 210
LDEXP_CONSTANT = 211
FREXP = 212
MODF = 213
FLOOR = 214
CEIL = 215
TRUNC = 216
DEGREES = 217
RADIANS = 218
CLIP = 219
ISREAL = 220
ISCOMPLEX = 221
REAL = 222
IMAG = 223
FIX = 224
I0 = 225
SINC = 226
NAN_TO_NUM = 227
ISCLOSE = 228
ISCLOSE_CONSTANT = 229
DIVMOD = 231
ANGLE = 232
SET_REAL = 233
SET_REAL_CONSTANT = 234
SET_IMAG = 235
SET_IMAG_CONSTANT = 236
TREE_ADD = 251
TREE_MULTIPLY = 252
CUMSUM = 301
CUMPROD = 302
PROD = 303
SUM = 304
MAX = 305
MIN = 306
ALL = 307
ANY = 308
MEAN_CHUNK = 309
MEAN_COMBINE = 310
MEAN = 311
ARGMAX = 312
ARGMAX_CHUNK = 313
ARGMAX_COMBINE = 314
ARGMIN = 315
ARGMIN_CHUNK = 316
ARGMIN_COMBINE = 317
NANSUM = 318
NANMAX = 319
NANMIN = 320
NANPROD = 321
NANMEAN = 322
NANMEAN_CHUNK = 323
NANARGMAX = 324
NANARGMAX_CHUNK = 325
NANARGMAX_COMBINE = 326
NANARGMIN = 327
NANARGMIN_CHUNK = 328
NANARGMIN_COMBINE = 329
COUNT_NONZERO = 330
MOMENT_CHUNK = 331
NANMOMENT_CHUNK = 332
MOMENT_COMBINE = 333
NANMOMENT_COMBINE = 334
MOMENT = 335
NANMOMENT = 336
VAR = 337
STD = 338
NANVAR = 339
NANSTD = 340
NANCUMSUM = 341
NANCUMPROD = 342
RESHAPE = 401
SLICE = 402
INDEX = 403
INDEXSETVALUE = 404
CONCATENATE = 405
RECHUNK = 406
ASTYPE = 407
TRANSPOSE = 408
SWAPAXES = 409
BROADCAST_TO = 410
STACK = 411
WHERE = 412
CHOOSE = 413
NONZERO = 414
ARGWHERE = 415
UNRAVEL_INDEX = 416
RAVEL_MULTI_INDEX = 417
ARRAY_SPLIT = 418
SQUEEZE = 419
DIGITIZE = 420
REPEAT = 421
COPYTO = 422
ISIN = 423
TENSORDOT = 501
DOT = 502
MATMUL = 503
CHOLESKY = 510
QR = 511
SVD = 512
LU = 513
SOLVE_TRIANGULAR = 520
NORM = 530
FFT = 601
IFFT = 602
FFT2 = 603
IFFT2 = 604
FFTN = 605
IFFTN = 606
RFFT = 607
IRFFT = 608
RFFT2 = 609
IRFFT2 = 610
RFFTN = 611
IRFFTN = 612
HFFT = 613
IHFFT = 614
FFTFREQ = 615
FFTFREQ_CHUNK = 616
RFFTFREQ = 617
FFTSHIFT = 618
IFFTSHIFT = 619
SPARSE_MATRIX_DATA_SOURCE = 701
DENSE_TO_SPARSE = 702
SPARSE_TO_DENSE = 703
FUSE = 801
ENTER = 901
LEAVE = 902
FIX_LATEST = 903
IF_ELSE = 904
NEXT_ITER = 905
TABLE_COO = 1003
STORE_COO = 1004
FETCH = 999999



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple

import numpy as np

from .core import Operand
from .. import opcodes as OperandDef
from ..serialize import TupleField, AnyField, Int64Field, BoolField, \
    KeyField, NDArrayField, StringField, Float64Field, ValueType
from ..compat import zip_longest
from ..utils import tokenize


class RandomOperand(Operand):
    _state = TupleField('state', on_serialize=lambda x: tuple(x) if x is not None else x,
                        on_deserialize=lambda x: State(x) if x is not None else x)

    @property
    def state(self):
        return getattr(self, '_state', None)

    def __setattr__(self, attr, value):
        if attr == '_state' and value is not None and not isinstance(value, State):
            value = State(value)
        super(RandomOperand, self).__setattr__(attr, value)

    @property
    def args(self):
        return [slot for slot in self.__slots__
                if slot not in set(RandomOperand.__slots__)]

    def update_key(self):
        args = tuple(getattr(self, k, None) for k in self.__slots__)
        if self.state is None:
            args += (np.random.random(),)
        self._key = tokenize(type(self), *args)


STATE = namedtuple('State', 'label keys pos has_gauss cached_gaussian')


class State(STATE):
    def __new__(cls, random_state=None, *args, **kwargs):
        if isinstance(random_state, np.random.RandomState) and not kwargs:
            return super(State, cls).__new__(cls, *random_state.get_state())
        elif isinstance(random_state, tuple) and not kwargs:
            return super(State, cls).__new__(cls, *random_state)
        elif args:
            return super(State, cls).__new__(cls, *((random_state,) + args))
        elif not kwargs:
            return super(State, cls).__new__(cls, *random_state)
        else:
            assert random_state is None
            return super(State, cls).__new__(cls, **kwargs)

    def __eq__(self, other):
        if not isinstance(other, tuple):
            return False

        for it, other_it in zip_longest(self, other):
            if isinstance(it, np.ndarray):
                if not np.array_equal(it, other_it):
                    return False
            elif it != other_it:
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def random_state(self):
        rs = np.random.RandomState()
        s = (self.label, self.keys, self.pos, self.has_gauss, self.cached_gaussian)
        rs.set_state(s)

        return rs


class SimpleRandomData(RandomOperand):
    __slots__ = ()


class Rand(SimpleRandomData):
    _op_type_ = OperandDef.RAND_RAND

    _size = TupleField('size', ValueType.int64)

    @property
    def size(self):
        return self._size


class Randn(SimpleRandomData):
    _op_type_ = OperandDef.RAND_RANDN

    _size = TupleField('size')

    @property
    def size(self):
        return self._size


class Randint(SimpleRandomData):
    __slots__ = '_low', '_high', '_density', '_size'
    _op_type_ = OperandDef.RAND_RANDINT

    _low = Int64Field('low')
    _high = Int64Field('high')
    _density = Float64Field('density')
    _size = TupleField('size', ValueType.int64)

    @property
    def low(self):
        return self._low

    @property
    def high(self):
        return self._high

    @property
    def density(self):
        return self._density

    @property
    def size(self):
        return self._size


class RandomIntegers(SimpleRandomData):
    __slots__ = '_low', '_high', '_size'
    _op_type_ = OperandDef.RAND_RANDOM_INTEGERS

    _low = Int64Field('low')
    _high = Int64Field('high')
    _size = TupleField('size', ValueType.int64)

    @property
    def low(self):
        return self._low

    @property
    def high(self):
        return self._high

    @property
    def size(self):
        return self._size


class RandomSample(SimpleRandomData):
    __slots__ = '_size',
    _op_type_ = OperandDef.RAND_RANDOM_SAMPLE

    _size = TupleField('size', ValueType.int64)

    @property
    def size(self):
        return self._size


class Random(SimpleRandomData):
    __slots__ = '_size',
    _op_type_ = OperandDef.RAND_RANDOM

    _size = TupleField('size', ValueType.int64)

    @property
    def size(self):
        return self._size


class Ranf(SimpleRandomData):
    __slots__ = '_size',
    _op_type_ = OperandDef.RAND_RANF

    _size = TupleField('size', ValueType.int64)

    @property
    def size(self):
        return self._size


class Sample(SimpleRandomData):
    __slots__ = '_size',
    _op_type_ = OperandDef.RAND_SAMPLE

    _size = TupleField('size', ValueType.int64)

    @property
    def size(self):
        return self._size


class Choice(SimpleRandomData):
    __slots__ = '_a', '_size', '_replace', '_p'
    _op_type_ = OperandDef.RAND_CHOICE

    _a = AnyField('a')
    _size = TupleField('size', ValueType.int64)
    _replace = BoolField('replace')
    _p = KeyField('p')

    @property
    def a(self):
        return self._a

    @property
    def size(self):
        return self._size

    @property
    def replace(self):
        return self._replace

    @property
    def p(self):
        return self._p


class Distribution(RandomOperand):
    __slots__ = ()


class Beta(Distribution):
    __slots__ = '_a', '_b', '_size'
    _op_type_ = OperandDef.RAND_BETA

    _a = AnyField('a')
    _b = AnyField('b')
    _size = TupleField('size', ValueType.int64)

    @property
    def a(self):
        return self._a

    @property
    def b(self):
        return self._b

    @property
    def size(self):
        return self._size


class Binomial(Distribution):
    __slots__ = '_n', '_p', '_size'
    _op_type_ = OperandDef.RAND_BINOMIAL

    _n = AnyField('n')
    _p = AnyField('p')
    _size = TupleField('size', ValueType.int64)

    @property
    def n(self):
        return self._n

    @property
    def p(self):
        return self._p

    @property
    def size(self):
        return self._size


class Chisquare(Distribution):
    __slots__ = '_df', '_size'
    _op_type_ = OperandDef.RAND_CHISQUARE

    _df = AnyField('df')
    _size = TupleField('size', ValueType.int64)

    @property
    def df(self):
        return self._df

    @property
    def size(self):
        return self._size


class Dirichlet(Distribution):
    __slots__ = '_alpha', '_size'
    _op_type_ = OperandDef.RAND_DIRICHLET

    _alpha = TupleField('alpha')
    _size = TupleField('size', ValueType.int64)

    @property
    def alpha(self):
        return self._alpha

    @property
    def size(self):
        return self._size


class Exponential(Distribution):
    __slots__ = '_scale', '_size'
    _op_type_ = OperandDef.RAND_EXPONENTIAL

    _scale = AnyField('scale')
    _size = TupleField('size', ValueType.int64)

    @property
    def scale(self):
        return self._scale

    @property
    def size(self):
        return self._size


class F(Distribution):
    __slots__ = '_dfnum', '_dfden', '_size'
    _op_type_ = OperandDef.RAND_F

    _dfnum = AnyField('dfnum')
    _dfden = AnyField('dfden')
    _size = TupleField('size', ValueType.int64)

    @property
    def dfnum(self):
        return self._dfnum

    @property
    def dfden(self):
        return self._dfden

    @property
    def size(self):
        return self._size


class Gamma(Distribution):
    __slots__ = '_shape', '_scale', '_size'
    _op_type_ = OperandDef.RAND_GAMMA

    _shape = AnyField('shape')
    _scale = AnyField('scale')
    _size = TupleField('size', ValueType.int64)

    @property
    def shape(self):
        return self._shape

    @property
    def scale(self):
        return self._scale

    @property
    def size(self):
        return self._size


class Geometric(Distribution):
    __slots__ = '_p', '_size'
    _op_type_ = OperandDef.RAND_GEOMETRIC

    _p = AnyField('p')
    _size = TupleField('size', ValueType.int64)

    @property
    def p(self):
        return self._p

    @property
    def size(self):
        return self._size


class Gumbel(Distribution):
    __slots__ = '_loc', '_scale', '_size'
    _op_type_ = OperandDef.RAND_GUMBEL

    _loc = AnyField('loc')
    _scale = AnyField('scale')
    _size = TupleField('size', ValueType.int64)

    @property
    def loc(self):
        return self._loc

    @property
    def scale(self):
        return self._scale

    @property
    def size(self):
        return self._size


class Hypergeometric(Distribution):
    __slots__ = '_ngood', '_nbad', '_nsample', '_size'
    _op_type_ = OperandDef.RAND_HYPERGEOMETRIC

    _ngood = AnyField('ngood')
    _nbad = AnyField('nbad')
    _nsample = AnyField('nsample')
    _size = TupleField('size', ValueType.int64)

    @property
    def ngood(self):
        return self._ngood

    @property
    def nbad(self):
        return self._nbad

    @property
    def nsample(self):
        return self._nsample

    @property
    def size(self):
        return self._size


class Laplace(Distribution):
    __slots__ = '_loc', '_scale', '_size'
    _op_type_ = OperandDef.RAND_LAPLACE

    _loc = AnyField('loc')
    _scale = AnyField('scale')
    _size = TupleField('size', ValueType.int64)

    @property
    def loc(self):
        return self._loc

    @property
    def scale(self):
        return self._scale

    @property
    def size(self):
        return self._size


class Logistic(Distribution):
    __slots__ = '_loc', '_scale', '_size'
    _op_type_ = OperandDef.RAND_LOGISTIC

    _loc = AnyField('loc')
    _scale = AnyField('scale')
    _size = TupleField('size', ValueType.int64)

    @property
    def loc(self):
        return self._loc

    @property
    def scale(self):
        return self._scale

    @property
    def size(self):
        return self._size


class Lognormal(Distribution):
    __slots__ = '_mean', '_sigma', '_size'
    _op_type_ = OperandDef.RAND_LOGNORMAL

    _mean = AnyField('mean')
    _sigma = AnyField('sigma')
    _size = TupleField('size', ValueType.int64)

    @property
    def mean(self):
        return self._mean

    @property
    def sigma(self):
        return self._sigma

    @property
    def size(self):
        return self._size


class Logseries(Distribution):
    __slots__ = '_p', '_size'
    _op_type_ = OperandDef.RAND_LOGSERIES

    _p = AnyField('p')
    _size = TupleField('size', ValueType.int64)

    @property
    def p(self):
        return self._p

    @property
    def size(self):
        return self._size


class Multinomial(Distribution):
    __slots__ = '_n', '_pvals', '_size'
    _op_type_ = OperandDef.RAND_MULTINOMIAL

    _n = Int64Field('n')
    _pvals = TupleField('pvals', ValueType.float64)
    _size = TupleField('size', ValueType.int64)

    @property
    def n(self):
        return self._n

    @property
    def pvals(self):
        return self._pvals

    @property
    def size(self):
        return self._size


class MultivariateNormal(Distribution):
    __slots__ = '_mean', '_cov', '_size', '_check_valid', '_tol'
    _op_type_ = OperandDef.RAND_MULTIVARIATE_NORMAL

    _mean = NDArrayField('mean')
    _cov = NDArrayField('cov')
    _size = TupleField('size', ValueType.int64)
    _check_valid = StringField('check_valid')
    _tol = Float64Field('tol')

    @property
    def mean(self):
        return self._mean

    @property
    def cov(self):
        return self._cov

    @property
    def size(self):
        return self._size

    @property
    def check_valid(self):
        return self._check_valid

    @property
    def tol(self):
        return self._tol


class NegativeBinomial(Distribution):
    __slots__ = '_n', '_p', '_size'
    _op_type_ = OperandDef.RAND_NEGATIVE_BINOMIAL

    _n = AnyField('n')
    _p = AnyField('p')
    _size = TupleField('size', ValueType.int64)

    @property
    def n(self):
        return self._n

    @property
    def p(self):
        return self._p

    @property
    def size(self):
        return self._size


class NoncentralChisquare(Distribution):
    __slots__ = '_df', '_nonc', '_size'
    _op_type_ = OperandDef.RAND_NONCENTRAL_CHISQURE

    _df = AnyField('df')
    _nonc = AnyField('nonc')
    _size = TupleField('size', ValueType.int64)

    @property
    def df(self):
        return self._df

    @property
    def nonc(self):
        return self._nonc

    @property
    def size(self):
        return self._size


class NoncentralF(Distribution):
    __slots__ = '_dfnum', '_dfden', '_nonc', '_size'
    _op_type_ = OperandDef.RAND_NONCENTRAL_F

    _dfnum = AnyField('dfnum')
    _dfden = AnyField('dfden')
    _nonc = AnyField('nonc')
    _size = TupleField('size', ValueType.int64)

    @property
    def dfnum(self):
        return self._dfnum

    @property
    def dfden(self):
        return self._dfden

    @property
    def nonc(self):
        return self._nonc

    @property
    def size(self):
        return self._size


class Normal(Distribution):
    __slots__ = '_loc', '_scale', '_size'
    _op_type_ = OperandDef.RAND_NORMAL

    _loc = AnyField('loc')
    _scale = AnyField('scale')
    _size = TupleField('size', ValueType.int64)

    @property
    def loc(self):
        return self._loc

    @property
    def scale(self):
        return self._scale

    @property
    def size(self):
        return self._size


class Pareto(Distribution):
    __slots__ = '_a', '_size'
    _op_type_ = OperandDef.RAND_PARETO

    _a = AnyField('a')
    _size = TupleField('size', ValueType.int64)

    @property
    def a(self):
        return self._a

    @property
    def size(self):
        return self._size


class Poisson(Distribution):
    __slots__ = '_lam', '_size'
    _op_type_ = OperandDef.RAND_POSSION

    _lam = AnyField('lam')
    _size = TupleField('size', ValueType.int64)

    @property
    def lam(self):
        return self._lam

    @property
    def size(self):
        return self._size


class RandomPower(Distribution):
    __slots__ = '_a', '_size'
    _op_type_ = OperandDef.RAND_POWER

    _a = AnyField('a')
    _size = TupleField('size', ValueType.int64)

    @property
    def a(self):
        return self._a

    @property
    def size(self):
        return self._size


class Rayleigh(Distribution):
    __slots__ = '_scale', '_size'
    _op_type_ = OperandDef.RAND_RAYLEIGH

    _scale = AnyField('scale')
    _size = TupleField('size', ValueType.int64)

    @property
    def scale(self):
        return self._scale

    @property
    def size(self):
        return self._size


class StandardCauchy(Distribution):
    __slots__ = '_size',
    _op_type_ = OperandDef.RAND_STANDARD_CAUCHY

    _size = TupleField('size', ValueType.int64)

    @property
    def size(self):
        return self._size


class StandardExponential(Distribution):
    __slots__ = '_size',
    _op_type_ = OperandDef.RAND_STANDARD_EXPONENTIAL

    _size = TupleField('size', ValueType.int64)

    @property
    def size(self):
        return self._size


class StandardGamma(Distribution):
    __slots__ = '_shape', '_size'
    _op_type_ = OperandDef.RAND_STANDARD_GAMMMA

    _shape = AnyField('shape')
    _size = TupleField('size', ValueType.int64)

    @property
    def shape(self):
        return self._shape

    @property
    def size(self):
        return self._size


class StandardNormal(Distribution):
    __slots__ = '_size',
    _op_type_ = OperandDef.RAND_STANDARD_NORMAL

    _size = TupleField('size', ValueType.int64)

    @property
    def size(self):
        return self._size


class StandardT(Distribution):
    __slots__ = '_df', '_size'
    _op_type_ = OperandDef.RAND_STANDARD_T

    _df = AnyField('df')
    _size = TupleField('size', ValueType.int64)

    @property
    def df(self):
        return self._df

    @property
    def size(self):
        return self._size


class Triangular(Distribution):
    __slots__ = '_left', '_mode', '_right', '_size'
    _op_type_ = OperandDef.RAND_TRIANGULAR

    _left = AnyField('left')
    _mode = AnyField('mode')
    _right = AnyField('right')
    _size = TupleField('size', ValueType.int64)

    @property
    def left(self):
        return self._left

    @property
    def mode(self):
        return self._mode

    @property
    def right(self):
        return self._right

    @property
    def size(self):
        return self._size


class Uniform(Distribution):
    __slots__ = '_low', '_high', '_size'
    _op_type_ = OperandDef.RAND_UNIFORM

    _low = AnyField('low')
    _high = AnyField('high')
    _size = TupleField('size', ValueType.int64)

    @property
    def low(self):
        return self._low

    @property
    def high(self):
        return self._high

    @property
    def size(self):
        return self._size


class Vonmises(Distribution):
    __slots__ = '_mu', '_kappa', '_size'
    _op_type_ = OperandDef.RAND_VONMISES

    _mu = AnyField('mu')
    _kappa = AnyField('kappa')
    _size = TupleField('size', ValueType.int64)

    @property
    def mu(self):
        return self._mu

    @property
    def kappa(self):
        return self._kappa

    @property
    def size(self):
        return self._size


class Wald(Distribution):
    __slots__ = '_mean', '_scale', '_size'
    _op_type_ = OperandDef.RAND_WALD

    _mean = AnyField('mean')
    _scale = AnyField('scale')
    _size = TupleField('size', ValueType.int64)

    @property
    def mean(self):
        return self._mean

    @property
    def scale(self):
        return self._scale

    @property
    def size(self):
        return self._size


class Weibull(Distribution):
    __slots__ = '_a', '_size'
    _op_type_ = OperandDef.RAND_WEIBULL

    _a = AnyField('a')
    _size = TupleField('size', ValueType.int64)

    @property
    def a(self):
        return self._a

    @property
    def size(self):
        return self._size


class Zipf(Distribution):
    __slots__ = '_a', '_size'
    _op_type_ = OperandDef.RAND_ZIPF

    _a = AnyField('a')
    _size = TupleField('size', ValueType.int64)

    @property
    def a(self):
        return self._a

    @property
    def size(self):
        return self._size



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import infer_dtype
from ..datasource import tensor as astensor
from .core import TensorBinOp, TensorConstant


class TensorLdexp(operands.Ldexp, TensorBinOp):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorLdexp, self).__init__(_casting=casting, _err=err,
                                          _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        return x1.issparse()

    @classmethod
    def constant_cls(cls):
        return TensorLdexpConstant


class TensorLdexpConstant(operands.LdexpConstant, TensorConstant):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorLdexpConstant, self).__init__(_casting=casting, _err=err,
                                                  _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        if hasattr(x1, 'issparse') and x1.issparse():
            return True
        return False


@infer_dtype(np.ldexp)
def ldexp(x1, x2, out=None, where=None, **kwargs):
    """
    Returns x1 * 2**x2, element-wise.

    The mantissas `x1` and twos exponents `x2` are used to construct
    floating point numbers ``x1 * 2**x2``.

    Parameters
    ----------
    x1 : array_like
        Tensor of multipliers.
    x2 : array_like, int
        Tensor of twos exponents.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : Tensor or scalar
        The result of ``x1 * 2**x2``.

    See Also
    --------
    frexp : Return (y1, y2) from ``x = y1 * 2**y2``, inverse to `ldexp`.

    Notes
    -----
    Complex dtypes are not supported, they will raise a TypeError.

    `ldexp` is useful as the inverse of `frexp`, if used by itself it is
    more clear to simply use the expression ``x1 * 2**x2``.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.ldexp(5, mt.arange(4)).execute()
    array([  5.,  10.,  20.,  40.], dtype=float32)

    >>> x = mt.arange(6)
    >>> mt.ldexp(*mt.frexp(x)).execute()
    array([ 0.,  1.,  2.,  3.,  4.,  5.])
    """
    x2_dtype = astensor(x2).dtype
    casting = kwargs.get('casting', 'safe')
    if not np.can_cast(x2_dtype, np.int64, casting=casting):
        raise TypeError("ufunc 'ldexp' not supported for the input types, "
                        "and the inputs could not be safely coerced to any supported types "
                        "according to the casting rule ''{0}''".format(casting))

    op = TensorLdexp(**kwargs)
    return op(x1, x2, out=out, where=where)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from .utils import WorkerActor

logger = logging.getLogger(__name__)


class ProcessHelperActor(WorkerActor):
    """
    Actor handling utils on every process
    """
    def __init__(self):
        super(ProcessHelperActor, self).__init__()
        self._dispatch_ref = None

    def post_create(self):
        from .dispatcher import DispatchActor

        super(ProcessHelperActor, self).post_create()
        self._dispatch_ref = self.promise_ref(DispatchActor.default_name())
        self._dispatch_ref.register_free_slot(self.uid, 'process_helper')

    def free_mkl_buffers(self):
        """
        Free MKL buffer
        """
        from ..lib.mkl_interface import mkl_free_buffers
        if mkl_free_buffers is None:
            return
        mkl_free_buffers()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
import time
import logging

import requests

from ..compat import six, TimeoutError
from ..serialize import dataserializer
from ..errors import ExecutionInterrupted
from ..graph import DirectedGraph

logger = logging.getLogger(__name__)


class Session(object):
    def __init__(self, endpoint, args=None):
        self._endpoint = endpoint
        self._args = args
        self._tensor_to_graph = dict()

        self._req_session = requests.Session()

        from requests.adapters import HTTPAdapter
        self._req_session.mount('http://stackoverflow.com', HTTPAdapter(max_retries=5))
        self._main()

    @property
    def session_id(self):
        return self._session_id

    def _main(self):
        resp = self._req_session.post(self._endpoint + '/api/session', self._args)
        if resp.status_code >= 400:
            raise SystemError('Failed to create mars session.')
        content = json.loads(resp.text)
        self._session_id = content['session_id']

    def run(self, *tensors, **kw):
        timeout = kw.pop('timeout', -1)
        compose = kw.pop('compose', True)
        wait = kw.pop('wait', True)
        if kw:
            raise TypeError('run got unexpected key arguments {0}'.format(', '.join(kw.keys())))

        graph = DirectedGraph()
        for t in tensors:
            graph = t.build_graph(graph=graph, tiled=False, compose=compose)
        targets = [t.key for t in tensors]

        targets_join = ','.join(targets)
        session_url = self._endpoint + '/api/session/' + self._session_id
        graph_json = graph.to_json()

        resp_json = self._submit_graph(graph_json, targets_join)
        graph_key = resp_json['graph_key']
        graph_url = session_url + '/graph/' + graph_key

        for t in tensors:
            self._tensor_to_graph[t.key] = graph_key

        if wait:
            exec_start_time = time.time()
            while timeout <= 0 or time.time() - exec_start_time <= timeout:
                try:
                    time.sleep(1)
                    try:
                        resp = self._req_session.get(graph_url)
                    except requests.ConnectionError as ex:
                        err_msg = str(ex)
                        if 'ConnectionResetError' in err_msg or 'Connection refused' in err_msg:
                            continue
                        raise
                    if resp.status_code == 504:
                        logging.debug('Gateway Time-out, try again')
                        continue
                    if resp.status_code >= 400:
                        raise SystemError('Failed to read task status. Code: %d, Reason: %s, Content:\n%s' %
                                          (resp.status_code, resp.reason, resp.text))
                    resp_json = json.loads(resp.text)
                    if resp_json['state'] in ('running', 'preparing'):
                        continue
                    elif resp_json['state'] == 'success':
                        break
                    elif resp_json['state'] == ('cancelled', 'cancelling'):
                        raise ExecutionInterrupted
                    elif resp_json['state'] == 'failed':
                        # TODO add traceback
                        if 'traceback' in resp_json:
                            traceback = resp_json['traceback']
                            if isinstance(traceback, list):
                                traceback = ''.join(str(s) for s in traceback)
                            raise SystemError('Graph execution failed.\nMessage: %s\nTraceback from server:\n%s' %
                                              (resp_json['msg'], traceback))
                        else:
                            raise SystemError('Graph execution failed with unknown reason.')
                    else:
                        raise SystemError('Unknown graph execution state %s' % resp_json['state'])
                except KeyboardInterrupt:
                    resp = self._req_session.delete(graph_url)
                    if resp.status_code >= 400:
                        raise SystemError('Failed to stop graph execution. Code: %d, Reason: %s, Content:\n%s' %
                                          (resp.status_code, resp.reason, resp.text))
            if 0 < timeout < time.time() - exec_start_time:
                raise TimeoutError
            data_list = []
            for tk in targets:
                resp = self._req_session.get(session_url + '/graph/' + graph_key + '/data/' + tk)
                if resp.status_code >= 400:
                    continue
                data_list.append(dataserializer.loads(resp.content))
            return data_list
        else:
            return graph_key

    def decref(self, *keys):
        session_url = self._endpoint + '/api/session/' + self._session_id
        for k in keys:
            if k not in self._tensor_to_graph:
                continue
            data_url = session_url + '/graph/%s/data/%s' % (self._tensor_to_graph[k], k)
            self._req_session.delete(data_url)

    def stop(self, graph_key):
        session_url = self._endpoint + '/api/session/' + self._session_id
        graph_url = session_url + '/graph/' + graph_key
        resp = self._req_session.delete(graph_url)
        if resp.status_code >= 400:
            raise SystemError('Failed to stop graph execution. Code: %d, Reason: %s, Content:\n%s' %
                              (resp.status_code, resp.reason, resp.text))

    def _submit_graph(self, graph_json, targets):
        session_url = self._endpoint + '/api/session/' + self._session_id
        resp = self._req_session.post(session_url + '/graph', dict(
            graph=json.dumps(graph_json),
            target=targets,
        ))
        if resp.status_code >= 400:
            resp_json = json.loads(resp.text)
            exc_info = base64.b64decode(resp_json['exc_info'])
            six.reraise(*exc_info)
        resp_json = json.loads(resp.text)
        return resp_json

    def close(self):
        self.decref(*list(self._tensor_to_graph.keys()))

        resp = self._req_session.delete(self._endpoint + '/api/session/' + self._session_id)
        if resp.status_code >= 400:
            raise SystemError('Failed to close mars session.')

    def check_service_ready(self, timeout=1):
        try:
            resp = self._req_session.get(self._endpoint + '/api', timeout=timeout)
        except (requests.ConnectionError, requests.Timeout):
            return False
        if resp.status_code >= 400:
            return False
        return True

    def count_workers(self):
        resp = self._req_session.get(self._endpoint + '/api/worker', timeout=1)
        return json.loads(resp.text)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
try:
    import scipy.fftpack as scifft
except ImportError:
    scifft = None

from ...operands import fft as fftop
from .array import get_array_module


def _get_fft_func(op, xp):
    fun_name = type(op).__name__.lower()[6:]  # all op starts with tensor
    if type(op) in (fftop.FFT, fftop.IFFT, fftop.FFT2, fftop.IFFT2,
                    fftop.FFTN, fftop.IFFTN):
        if xp is np and scifft and op.norm is None:
            def f(*args, **kwargs):
                kwargs.pop('norm', None)
                if 's' in kwargs:
                    kwargs['shape'] = kwargs.pop('s', None)
                return getattr(scifft, fun_name)(*args, **kwargs)
            return f
        else:
            return getattr(xp.fft, fun_name)
    else:
        return getattr(xp.fft, fun_name)


def _fft(ctx, chunk):
    a = ctx[chunk.inputs[0].key]
    xp = get_array_module(a)
    fun = _get_fft_func(chunk.op, xp)
    res = fun(a, n=chunk.op.n, axis=chunk.op.axis, norm=chunk.op.norm)
    if res.dtype != chunk.op.dtype:
        res = res.astype(chunk.op.dtype)
    ctx[chunk.key] = res


def _fftn(ctx, chunk):
    a = ctx[chunk.inputs[0].key]
    xp = get_array_module(a)
    fun = _get_fft_func(chunk.op, xp)
    res = fun(a, s=chunk.op.shape, axes=chunk.op.axes, norm=chunk.op.norm)
    if res.dtype != chunk.op.dtype:
        res = res.astype(chunk.op.dtype)
    ctx[chunk.key] = res


def _fftfreq_chunk(ctx, chunk):
    n, d = chunk.op.n, chunk.op.d
    x = ctx[chunk.inputs[0].key].copy()
    x[x >= (n + 1) // 2] -= n
    x /= n * d
    ctx[chunk.key] = x


def register_fft_handler():
    from ... import operands
    from .core import register

    register(operands.StandardFFT, _fft)
    register(operands.StandardFFTN, _fftn)
    register(operands.RealFFT, _fft)
    register(operands.RealFFTN, _fftn)
    register(operands.HermitianFFT, _fft)
    register(operands.FFTFreqChunk, _fftfreq_chunk)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ....operands import fft as fftop
from ..datasource import tensor as astensor
from .core import TensorFFTMixin, validate_fft


class TensorIRFFT(fftop.IRFFT, TensorFFTMixin):
    def __init__(self, n=None, axis=-1, norm=None, dtype=None, **kw):
        super(TensorIRFFT, self).__init__(_n=n, _axis=axis, _norm=norm,
                                          _dtype=dtype, **kw)

    @classmethod
    def _get_shape(cls, op, shape):
        new_shape = list(shape)
        if op.n is not None:
            new_shape[op.axis] = op.n
        new_shape[op.axis] = 2 * (new_shape[op.axis] - 1)
        return tuple(new_shape)

    def _set_inputs(self, inputs):
        super(TensorIRFFT, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, a):
        shape = self._get_shape(self, a.shape)
        return self.new_tensor([a], shape)


def irfft(a, n=None, axis=-1, norm=None):
    """
    Compute the inverse of the n-point DFT for real input.

    This function computes the inverse of the one-dimensional *n*-point
    discrete Fourier Transform of real input computed by `rfft`.
    In other words, ``irfft(rfft(a), len(a)) == a`` to within numerical
    accuracy. (See Notes below for why ``len(a)`` is necessary here.)

    The input is expected to be in the form returned by `rfft`, i.e. the
    real zero-frequency term followed by the complex positive frequency terms
    in order of increasing frequency.  Since the discrete Fourier Transform of
    real input is Hermitian-symmetric, the negative frequency terms are taken
    to be the complex conjugates of the corresponding positive frequency terms.

    Parameters
    ----------
    a : array_like
        The input tensor.
    n : int, optional
        Length of the transformed axis of the output.
        For `n` output points, ``n//2+1`` input points are necessary.  If the
        input is longer than this, it is cropped.  If it is shorter than this,
        it is padded with zeros.  If `n` is not given, it is determined from
        the length of the input along the axis specified by `axis`.
    axis : int, optional
        Axis over which to compute the inverse FFT. If not given, the last
        axis is used.
    norm : {None, "ortho"}, optional
        Normalization mode (see `mt.fft`). Default is None.

    Returns
    -------
    out : Tensor
        The truncated or zero-padded input, transformed along the axis
        indicated by `axis`, or the last one if `axis` is not specified.
        The length of the transformed axis is `n`, or, if `n` is not given,
        ``2*(m-1)`` where ``m`` is the length of the transformed axis of the
        input. To get an odd number of output points, `n` must be specified.

    Raises
    ------
    IndexError
        If `axis` is larger than the last axis of `a`.

    See Also
    --------
    mt.fft : For definition of the DFT and conventions used.
    rfft : The one-dimensional FFT of real input, of which `irfft` is inverse.
    fft : The one-dimensional FFT.
    irfft2 : The inverse of the two-dimensional FFT of real input.
    irfftn : The inverse of the *n*-dimensional FFT of real input.

    Notes
    -----
    Returns the real valued `n`-point inverse discrete Fourier transform
    of `a`, where `a` contains the non-negative frequency terms of a
    Hermitian-symmetric sequence. `n` is the length of the result, not the
    input.

    If you specify an `n` such that `a` must be zero-padded or truncated, the
    extra/removed values will be added/removed at high frequencies. One can
    thus resample a series to `m` points via Fourier interpolation by:
    ``a_resamp = irfft(rfft(a), m)``.

    Examples
    --------
    >>> import mars.tenosr as mt

    >>> mt.fft.ifft([1, -1j, -1, 1j]).execute()
    array([ 0.+0.j,  1.+0.j,  0.+0.j,  0.+0.j])
    >>> mt.fft.irfft([1, -1j, -1]).execute()
    array([ 0.,  1.,  0.,  0.])

    Notice how the last term in the input to the ordinary `ifft` is the
    complex conjugate of the second term, and the output has zero imaginary
    part everywhere.  When calling `irfft`, the negative frequencies are not
    specified, and the output array is purely real.

    """
    a = astensor(a)
    validate_fft(a, axis=axis, norm=norm)
    op = TensorIRFFT(n=n, axis=axis, norm=norm, dtype=np.dtype(np.float_))
    return op(a)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time

from .. import resource
from ..compat import six, OrderedDict3
from ..utils import log_unhandled
from .utils import WorkerActor

logger = logging.getLogger(__name__)


class QuotaActor(WorkerActor):
    """
    Actor handling quota request and assignment
    """
    def __init__(self, total_size):
        super(QuotaActor, self).__init__()
        self._status_ref = None

        self._requests = OrderedDict3()

        self._total_size = total_size
        self._allocations = dict()
        self._allocated_size = 0

        self._proc_sizes = dict()
        self._total_proc = 0

        self._hold_sizes = dict()
        self._total_hold = 0

    def post_create(self):
        from .status import StatusActor

        super(QuotaActor, self).post_create()

        status_ref = self.ctx.actor_ref(StatusActor.default_name())
        if self.ctx.has_actor(status_ref):
            self._status_ref = status_ref

    def _has_space(self, delta):
        return self._allocated_size + delta <= self._total_size

    @log_unhandled
    def request_batch_quota(self, batch, callback):
        """
        Request for resources in a batch
        :param batch: the request dict in form {request_key: request_size, ...}
        :param callback: promise callback
        :return: if request is returned immediately, return True, otherwise False
        """
        all_allocated = True
        # check if the request is already allocated
        for key, size in batch.items():
            if key not in self._allocations or size > self._allocations.get(key):
                all_allocated = False
                break

        # if all requested and allocation can still be applied, apply directly
        if all_allocated and self._has_space(0):
            if callback is not None:
                self.tell_promise(callback)
            return True

        logger.debug('Receive batch quota request %r on %s. Allocated: %s. Total size: %s',
                     batch, self.uid, self._allocated_size, self._total_size)
        sorted_req = sorted(batch.items(), key=lambda tp: tp[0])
        keys = tuple(tp[0] for tp in sorted_req)
        values = tuple(tp[1] for tp in sorted_req)
        delta = sum(v - self._allocations.get(k, 0) for k, v in batch.items())
        # make allocated requests the highest priority to be allocated
        return self._request_quota(keys, values, delta, callback, make_first=all_allocated)

    @log_unhandled
    def request_quota(self, key, quota_size, callback):
        """
        Request for resource
        :param key: request key
        :param quota_size: size of request quota
        :param callback: promise callback
        :return: if request is returned immediately, return True, otherwise False
        """
        logger.debug('Receive quota request for key %s on %s. Allocated: %s. Total size: %s',
                     key, self.uid, self._allocated_size, self._total_size)
        quota_size = int(quota_size)
        make_first = False
        # check if the request is already allocated
        if key in self._allocations:
            old_size = self._allocations[key]
            # if all requested and allocation can still be applied, apply directly
            if old_size <= quota_size and self._has_space(0):
                self.tell_promise(callback)
                return True
            else:
                # make allocated requests the highest priority to be allocated
                make_first = True
        else:
            old_size = 0
        return self._request_quota(key, quota_size, quota_size - old_size, callback,
                                   make_first=make_first)

    def _request_quota(self, keys, quota_sizes, delta, callback, make_first=False):
        """
        Actually process requests
        :param keys: request keys
        :param quota_sizes: request sizes
        :param delta: increase of allocate size
        :param callback: promise callback
        :param make_first: whether to move request keys to the highest priority
        :return: if request is returned immediately, return True, otherwise False
        """
        if keys in self._requests:
            # already in request queue, store callback and quit
            self._requests[keys][-1].append(callback)
            if make_first:
                self._requests.move_to_end(keys, False)
            return False

        if not isinstance(keys, tuple) and self._allocations.get(keys, 0) >= quota_sizes:
            # already allocated, inform and quit
            self.tell_promise(callback)
            return True

        if self._has_space(delta):
            if not self._requests:
                # if no previous requests, we can apply directly
                logger.debug('Quota request met for key %r on %s. Allocated: %s. Total size: %s',
                             keys, self.uid, self._allocated_size, self._total_size)
                self.apply_allocation(keys, quota_sizes)
                if callback:
                    self.tell_promise(callback)
            else:
                # otherwise, previous requests are satisfied first
                logger.debug('Quota request queued for key %r on %s. Allocated: %s. Total size: %s',
                             keys, self.uid, self._allocated_size, self._total_size)
                if keys not in self._requests:
                    self._requests[keys] = (quota_sizes, delta, time.time(), [])
                    if make_first:
                        self._requests.move_to_end(keys, False)
                if callback:
                    self._requests[keys][-1].append(callback)
                    if make_first:
                        self._requests.move_to_end(keys, False)

            self._process_requests()
            return True
        else:
            # current free space cannot satisfy the request, the request is queued
            logger.debug('Quota request unmet for key %r on %s. Allocated: %s. Total size: %s',
                         keys, self.uid, self._allocated_size, self._total_size)
            if keys not in self._requests:
                self._requests[keys] = (quota_sizes, delta, time.time(), [])
                if make_first:
                    self._requests.move_to_end(keys, False)
            if callback:
                self._requests[keys][-1].append(callback)
                if make_first:
                    self._requests.move_to_end(keys, False)
            return False

    @log_unhandled
    def cancel_requests(self, keys, reject_exc=None):
        """
        Cancel a request if it is not assigned
        :param keys: request keys
        :param reject_exc: the exception to pass to the original callbacks
        """
        if isinstance(keys, six.string_types):
            keys = [keys]
        for k in keys:
            if k in self._requests:
                if reject_exc:
                    for cb in self._requests[keys][-1]:
                        self.tell_promise(cb, *reject_exc, **dict(_accept=False))
                del self._requests[k]

    @log_unhandled
    def process_quota(self, key):
        """
        Mark request quota as being processed
        :param key: request key
        """
        if key not in self._allocations:
            return
        alloc_size = self._allocations[key]
        self._total_proc += alloc_size - self._proc_sizes.get(key, 0)
        self._proc_sizes[key] = alloc_size

    @log_unhandled
    def hold_quota(self, key):
        """
        Mark request quota as already been hold
        :param key: request key
        """
        if key not in self._allocations:
            return
        alloc_size = self._allocations[key]
        self._total_hold += alloc_size - self._hold_sizes.get(key, 0)
        self._hold_sizes[key] = alloc_size

        if key in self._proc_sizes:
            self._total_proc -= self._proc_sizes[key]
            del self._proc_sizes[key]

    @log_unhandled
    def release_quota(self, key):
        """
        Release allocated quota
        :param key: request key
        """
        if key not in self._allocations:
            return
        alloc_size = self._allocations[key]
        self._allocated_size -= alloc_size
        del self._allocations[key]

        if key in self._proc_sizes:
            self._total_proc -= self._proc_sizes[key]
            del self._proc_sizes[key]

        if key in self._hold_sizes:
            self._total_hold -= self._hold_sizes[key]
            del self._hold_sizes[key]

        self._process_requests()
        logger.debug('Quota key %s released on %s. Allocated: %s. Total size: %s',
                     key, self.uid, self._allocated_size, self._total_size)

    @log_unhandled
    def release_quotas(self, keys):
        """
        Release allocated quota in batch
        :param keys: request keys
        """
        for k in keys:
            self.release_quota(k)

    def dump_state(self):
        logger.debug('State dump of %s:\nAllocations:%r\nRequests:%r',
                     self.uid, self._allocations, self._requests)

    def get_allocated_size(self):
        # get total allocated size, for debug purpose
        return self._allocated_size

    @log_unhandled
    def apply_allocation(self, key, quota_size, handle_shrink=True):
        """
        Accept a request
        :param key: request key
        :param quota_size: requested quota size
        :param handle_shrink: if True and the quota size less than the original, process requests in the queue
        """
        if isinstance(key, tuple):
            for k, s in zip(key, quota_size):
                self.apply_allocation(k, s, handle_shrink=handle_shrink)
            return
        quota_size = int(quota_size)
        old_size = self._allocations.get(key, 0)
        self._allocated_size += quota_size - old_size
        self._allocations[key] = quota_size
        if key in self._proc_sizes:
            self._total_proc += quota_size - self._proc_sizes[key]
            self._proc_sizes[key] = quota_size
        if key in self._hold_sizes:
            self._total_hold += quota_size - self._hold_sizes[key]
            self._hold_sizes[key] = quota_size
        logger.debug('Quota key %s applied on %s. Allocated: %s. Total size: %s',
                     key, self.uid, self._allocated_size, self._total_size)
        if handle_shrink and quota_size < old_size:
            self._process_requests()

    @log_unhandled
    def _process_requests(self):
        """
        Process quota requests in the queue
        """
        removed = []
        for k, req in six.iteritems(self._requests):
            req_size, delta, req_time, callbacks = req
            if self._has_space(delta):
                self.apply_allocation(k, req_size, handle_shrink=False)
                for cb in callbacks:
                    self.tell_promise(cb)
                if self._status_ref:
                    self._status_ref.update_mean_stats(
                        'wait_time.' + self.uid.replace('Actor', ''), time.time() - req_time,
                        _tell=True, _wait=False)
                removed.append(k)
            else:
                # Quota left cannot satisfy the next request, we quit
                break
        for k in removed:
            del self._requests[k]


class MemQuotaActor(QuotaActor):
    """
    Actor handling worker memory quota
    """
    def __init__(self, total_size, overall_size=None):
        super(MemQuotaActor, self).__init__(total_size)
        self._overall_size = overall_size or total_size
        self._last_memory_available = 0

        self._dispatch_ref = None

    def post_create(self):
        from .dispatcher import DispatchActor

        super(MemQuotaActor, self).post_create()
        self.update_mem_stats()
        self._dispatch_ref = self.promise_ref(DispatchActor.default_name())

        if self._status_ref:
            self._status_ref.set_mem_quota_allocations(
                dict(allocated=self._allocated_size, hold=self._total_hold, total=self._total_size),
                _tell=True, _wait=False)

    def update_mem_stats(self):
        """
        Refresh memory usage
        """
        cur_mem_available = resource.virtual_memory().available
        if cur_mem_available > self._last_memory_available:
            # memory usage reduced: try reallocate existing requests
            self._process_requests()
        self._last_memory_available = cur_mem_available
        self.ref().update_mem_stats(_tell=True, _delay=10)

    def _has_space(self, delta):
        mem_stats = resource.virtual_memory()
        # calc available physical memory
        available_size = mem_stats.available - min(0, mem_stats.total - self._overall_size) \
                         - self._total_proc
        if delta >= available_size:
            logger.warning('%s met hard memory limitation: request %d, available %d, hard limit %d',
                           self.uid, delta, available_size, self._overall_size)
            for slot in self._dispatch_ref.get_slots('process_helper'):
                self.ctx.actor_ref(slot).free_mkl_buffers(_tell=True, _wait=False)
            return False
        return super(MemQuotaActor, self)._has_space(delta)

    def apply_allocation(self, key, quota_size, handle_shrink=True):
        ret = super(MemQuotaActor, self).apply_allocation(
            key, quota_size, handle_shrink=handle_shrink)
        if self._status_ref:
            self._status_ref.set_mem_quota_allocations(
                dict(allocated=self._allocated_size, hold=self._total_hold, total=self._total_size),
                _tell=True, _wait=False)
        return ret

    def release_quota(self, key):
        ret = super(MemQuotaActor, self).release_quota(key)
        if self._status_ref:
            self._status_ref.set_mem_quota_allocations(
                dict(allocated=self._allocated_size, total=self._total_size),
                _tell=True, _wait=False)
        return ret

    def dump_keys(self):
        return list(self._allocations.keys())



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
import uuid
from functools import partial

from .spill import read_spill_file, write_spill_file, spill_exists
from .utils import WorkerActor
from .. import promise
from ..config import options
from ..errors import *
from ..utils import deserialize_graph, log_unhandled, calc_data_size
from ..compat import six, OrderedDict3
from ..tensor.execution.core import Executor

logger = logging.getLogger(__name__)

_calc_result_cache = OrderedDict3()


class InProcessCacheActor(WorkerActor):
    """
    Actor managing calculation result in rss memory
    """
    def __init__(self):
        super(InProcessCacheActor, self).__init__()
        self._chunk_holder_ref = None
        self._mem_quota_ref = None

        self._spill_dump_pool = None

    def post_create(self):
        from .chunkholder import ChunkHolderActor
        from .quota import MemQuotaActor

        super(InProcessCacheActor, self).post_create()
        self._chunk_holder_ref = self.promise_ref(ChunkHolderActor.default_name())
        self._mem_quota_ref = self.promise_ref(MemQuotaActor.default_name())
        if options.worker.spill_directory:
            self._spill_dump_pool = self.ctx.threadpool(len(options.worker.spill_directory))

    @promise.reject_on_exception
    @log_unhandled
    def dump_cache(self, keys, callback):
        """
        Dump data in rss memory into shared cache
        """
        @log_unhandled
        def _try_put_chunk(session_id, chunk_key, data_size):
            logger.debug('Try putting %s into shared cache.', chunk_key)
            try:
                if chunk_key not in _calc_result_cache:
                    if not self._chunk_store.contains(session_id, chunk_key):
                        raise KeyError('Data key %s not found in inproc cache', chunk_key)
                    return

                ref = None
                try:
                    ref = self._chunk_store.put(session_id, chunk_key, _calc_result_cache[chunk_key][1])
                    del _calc_result_cache[chunk_key]
                    self._mem_quota_ref.release_quota(chunk_key, _tell=True)

                    self._chunk_holder_ref.register_chunk(session_id, chunk_key)
                    self._kv_store_ref.write('/sessions/%s/chunks/%s/workers/%s' % (session_id, chunk_key, self.address), '')
                finally:
                    del ref

            except StoreFull:
                # if we cannot put data into shared cache, we store it into spill directly
                self._chunk_holder_ref.spill_size(data_size, _tell=True)
                _put_spill_directly(session_id, chunk_key, data_size)

        @log_unhandled
        def _put_spill_directly(session_id, chunk_key, data_size, *_):
            if self._spill_dump_pool is None:
                raise SpillNotConfigured

            logger.debug('Writing data %s directly into spill.', chunk_key)
            self._spill_dump_pool.submit(write_spill_file, chunk_key, _calc_result_cache[chunk_key][1]).result()

            del _calc_result_cache[chunk_key]
            self._mem_quota_ref.release_quota(chunk_key, _tell=True)

            chunk_path = '/sessions/%s/chunks/%s' % (session_id, chunk_key)
            futures = [
                self._kv_store_ref.write(chunk_path + '/data_size', data_size, _wait=False),
                self._kv_store_ref.write(chunk_path + '/workers/%s' % self.address, '', _wait=False),
            ]
            [f.result() for f in futures]

        promises = []
        for k in keys:
            session_id, value = _calc_result_cache[k]
            data_size = calc_data_size(value)
            del value
            promises.append(
                promise.Promise(done=True).then(partial(_try_put_chunk, session_id, k, data_size))
            )
        promise.all_(promises).then(lambda *_: self.tell_promise(callback)) \
            .catch(lambda *exc: self.tell_promise(callback, *exc, **dict(_accept=False)))

    @log_unhandled
    def remove_cache(self, keys):
        """
        Remove data from cache
        """
        for k in keys:
            del _calc_result_cache[k]


class CpuCalcActor(WorkerActor):
    def __init__(self):
        super(CpuCalcActor, self).__init__()
        self._mem_quota_ref = None
        self._inproc_cache_ref = None
        self._dispatch_ref = None
        self._status_ref = None

        self._execution_pool = None
        self._spill_load_pool = None

    def post_create(self):
        from .quota import MemQuotaActor
        from .dispatcher import DispatchActor
        from .status import StatusActor

        super(CpuCalcActor, self).post_create()
        if isinstance(self.uid, six.string_types) and ':' in self.uid:
            uid_parts = self.uid.split(':')
            inproc_uid = 'w:' + uid_parts[1] + ':inproc-cache-' + str(uuid.uuid4())
        else:
            inproc_uid = None
        self._inproc_cache_ref = self.promise_ref(self.ctx.create_actor(InProcessCacheActor, uid=inproc_uid))
        self._mem_quota_ref = self.promise_ref(MemQuotaActor.default_name())

        self._dispatch_ref = self.promise_ref(DispatchActor.default_name())
        self._dispatch_ref.register_free_slot(self.uid, 'cpu')

        self._status_ref = self.ctx.actor_ref(StatusActor.default_name())
        if not self.ctx.has_actor(self._status_ref):
            self._status_ref = None

        self._execution_pool = self.ctx.threadpool(1)
        if options.worker.spill_directory:
            self._spill_load_pool = self.ctx.threadpool(len(options.worker.spill_directory))

    @staticmethod
    def _build_load_key(graph_key, chunk_key):
        return '%s_load_memory_%s' % (graph_key, chunk_key)

    @promise.reject_on_exception
    @log_unhandled
    def calc(self, session_id, ser_graph, targets, callback):
        """
        Do actual calculation. This method should be called when all data
        is available (i.e., either in shared cache or in memory)
        :param session_id: session id
        :param ser_graph: serialized executable graph
        :param targets: keys of target chunks
        :param callback: promise callback, returns the uid of InProcessCacheActor
        """
        from ..tensor.expressions.datasource import TensorFetchChunk
        graph = deserialize_graph(ser_graph)

        ops = set(c.op for c in graph if not isinstance(c.op, TensorFetchChunk))
        op_name = '_'.join(sorted([type(o).__name__ for o in ops]))
        op_key = '_'.join(o.key for o in ops)

        try:
            context_dict = dict()
            comp_nodes = []
            absent_keys = []
            spill_load_futures = dict()
            for chunk in graph.iter_nodes():
                try:
                    # try load chunk from shared cache
                    if isinstance(chunk.op, TensorFetchChunk):
                        context_dict[chunk.key] = self._chunk_store.get(session_id, chunk.key)
                        self._mem_quota_ref.release_quota(self._build_load_key(op_key, chunk.key))
                    else:
                        comp_nodes.append(chunk.op.key)
                except KeyError:
                    # chunk not in shared cache, we load it from spill directly
                    if self._spill_load_pool is not None and spill_exists(chunk.key):
                        logger.debug('Load chunk %s directly from spill', chunk.key)
                        self._mem_quota_ref.process_quota(self._build_load_key(op_key, chunk.key))
                        spill_load_futures[chunk.key] = self._spill_load_pool.submit(read_spill_file, chunk.key)
                    else:
                        absent_keys.append(chunk.key)
            if absent_keys:
                raise ObjectNotInPlasma(absent_keys)

            # collect results from greenlets
            if spill_load_futures:
                for k, future in spill_load_futures.items():
                    context_dict[k] = future.result()
                    self._mem_quota_ref.hold_quota(self._build_load_key(op_key, k))
                spill_load_futures.clear()

            logger.debug('Start calculating operand %r.', comp_nodes)

            start_time = time.time()

            # mark targets as processing
            target_keys = [k for k in targets if not self._chunk_store.contains(session_id, k)]
            [self._mem_quota_ref.process_quota(k) for k in target_keys]

            # start actual execution
            executor = Executor(storage=context_dict)
            results = self._execution_pool.submit(executor.execute_graph, graph, targets).result()

            for k in list(context_dict.keys()):
                del context_dict[k]
                self._mem_quota_ref.release_quota(self._build_load_key(op_key, k))

            end_time = time.time()

            [self._mem_quota_ref.hold_quota(k) for k in target_keys]

            # adjust sizes in allocation
            save_sizes = dict()
            for k, v in zip(targets, results):
                if not self._chunk_store.contains(session_id, k):
                    _calc_result_cache[k] = (session_id, v)
                    save_sizes[k] = calc_data_size(v)
                    self._mem_quota_ref.apply_allocation(k, save_sizes[k])

            if self._status_ref:
                self._status_ref.update_mean_stats(
                    'calc_speed.' + op_name, sum(save_sizes.values()) * 1.0 / (end_time - start_time),
                    _tell=True, _wait=False)

            logger.debug('Finish calculating operand %r.', comp_nodes)
            self.tell_promise(callback, self._inproc_cache_ref.uid, save_sizes)
            self._dispatch_ref.register_free_slot(self.uid, 'cpu', _tell=True)
        except:
            self._dispatch_ref.register_free_slot(self.uid, 'cpu', _tell=True)
            raise



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import logging.config
import itertools
import platform
import struct
import warnings
import os
try:
    import xml.etree.cElementTree as ElementTree
except ImportError:
    import xml.etree.ElementTree as ElementTree
try:
    ElementTreeParseError = getattr(ElementTree, 'ParseError')
except AttributeError:
    ElementTreeParseError = getattr(ElementTree, 'XMLParserError')
from unicodedata import east_asian_width

from ..lib import six

PY27 = six.PY2 and sys.version_info[1] == 7
PYPY = platform.python_implementation().lower() == 'pypy'

SEEK_SET = 0
SEEK_CUR = 1
SEEK_END = 2

# Definition of East Asian Width
# http://unicode.org/reports/tr11/
# Ambiguous width can be changed by option
_EAW_MAP = {'Na': 1, 'N': 1, 'W': 2, 'F': 2, 'H': 1}

import decimal
DECIMAL_TYPES = [decimal.Decimal, ]

import json  # don't remove

if six.PY3:
    lrange = lambda *x: list(range(*x))
    lzip = lambda *x: list(zip(*x))
    lkeys = lambda x: list(x.keys())
    lvalues = lambda x: list(x.values())
    litems = lambda x: list(x.items())
    lmap = lambda *x: list(map(*x))

    irange = range
    izip = zip

    long_type = int

    import io
    StringIO = io.StringIO
    BytesIO = io.BytesIO

    if PY27:
        from ..lib import enum
    else:
        import enum

    if PY27:
        try:
            import cdecimal as decimal

            DECIMAL_TYPES.append(decimal.Decimal)
        except ImportError:
            import decimal
    else:
        import decimal

    from collections import OrderedDict

    OrderedDict3 = OrderedDict

    def u(s):
        return s

    def strlen(data, encoding=None):
        # encoding is for compat with PY2
        return len(data)

    def east_asian_len(data, encoding=None, ambiguous_width=1):
        """
        Calculate display width considering unicode East Asian Width
        """
        if isinstance(data, six.text_type):
            return sum([_EAW_MAP.get(east_asian_width(c), ambiguous_width) for c in data])
        else:
            return len(data)

    dictconfig = lambda config: logging.config.dictConfig(config)

    import builtins
    from concurrent import futures  # don't remove

    from datetime import timedelta
    total_seconds = timedelta.total_seconds

    import functools as functools32

    def np_getbuffer(n):
        return memoryview(n)

    BrokenPipeError = BrokenPipeError
    ConnectionResetError = ConnectionResetError
    TimeoutError = TimeoutError

    from itertools import accumulate
else:
    lrange = range
    lzip = zip
    lkeys = lambda x: x.keys()
    lvalues = lambda x: x.values()
    litems = lambda x: x.items()
    lmap = map

    irange = xrange  # noqa F821
    izip = itertools.izip

    long_type = long  # noqa F821

    from ..lib import enum

    try:
        import cdecimal as decimal
        DECIMAL_TYPES.append(decimal.Decimal)
    except ImportError:
        import decimal

    try:
        import cStringIO as StringIO
    except ImportError:
        import StringIO
    StringIO = BytesIO = StringIO.StringIO

    def u(s):
        return unicode(s, "unicode_escape")  # noqa F821

    def strlen(data, encoding=None):
        try:
            data = data.decode(encoding)
        except UnicodeError:
            pass
        return len(data)

    def east_asian_len(data, encoding=None, ambiguous_width=1):
        """
        Calculate display width considering unicode East Asian Width
        """
        if isinstance(data, six.text_type):
            try:
                data = data.decode(encoding)
            except UnicodeError:
                pass
            return sum([_EAW_MAP.get(east_asian_width(c), ambiguous_width) for c in data])
        else:
            return len(data)

    from collections import OrderedDict

    dictconfig = lambda config: logging.config.dictConfig(config)

    from datetime import timedelta
    total_seconds = timedelta.total_seconds

    import __builtin__ as builtins  # don't remove
    from ..lib import futures  # don't remove
    from ..lib.functools32.functools32 import OrderedDict as OrderedDict3

    from ..lib import functools32  # don't remove

    def np_getbuffer(n):
        import numpy as np
        return np.getbuffer(n)


    class TimeoutError(Exception):
        pass


    class BrokenPipeError(Exception):
        pass


    class ConnectionResetError(Exception):
        pass


    def accumulate(iterable, func=lambda a, b: a + b):
        'Return running totals'
        # accumulate([1,2,3,4,5]) --> 1 3 6 10 15
        # accumulate([1,2,3,4,5], operator.mul) --> 1 2 6 24 120
        it = iter(iterable)
        try:
            total = next(it)
        except StopIteration:
            return
        yield total
        for element in it:
            total = func(total, element)
            yield total

if six.PY3:
    from contextlib import suppress
else:
    from contextlib import contextmanager

    @contextmanager
    def suppress(*exceptions):
        try:
            yield
        except exceptions:
            pass

Enum = enum.Enum
DECIMAL_TYPES = tuple(DECIMAL_TYPES)
Decimal = decimal.Decimal

if sys.version_info.major < 3:
    # Due to a bug in python 2.7 Queue.get, if a timeout isn't specified then
    # `Queue.get` can't be interrupted. A workaround is to specify an extremely
    # long timeout, which then allows it to be interrupted.
    # For more information see: https://bugs.python.org/issue1360
    def queue_get(q):
        return q.get(block=True, timeout=(365 * 24 * 60 * 60))

elif os.name == 'nt':
    # Python 3 windows Queue.get also doesn't handle interrupts properly. To
    # workaround this we poll at a sufficiently large interval that it
    # shouldn't affect performance, but small enough that users trying to kill
    # an application shouldn't care.
    def queue_get(q):
        while True:
            try:
                return q.get(block=True, timeout=0.1)
            except Empty:
                pass
else:
    def queue_get(q):
        return q.get()

from ..lib.lib_utils import isvalidattr, dir2, raise_exc, getargspec, getfullargspec

from ..lib.six.moves import reduce, zip_longest
from ..lib.six.moves import reload_module
from ..lib.six.moves.queue import Queue, Empty, PriorityQueue
from ..lib.six.moves.urllib.request import urlretrieve
from ..lib.six.moves import cPickle as pickle
from ..lib.six.moves.urllib.parse import urlencode, urlparse, unquote, quote, quote_plus, parse_qsl
from ..lib.six.moves import configparser as ConfigParser


try:
    import pytz
    utc = pytz.utc
    FixedOffset = pytz._FixedOffset
except ImportError:
    import datetime
    _ZERO_TIMEDELTA = datetime.timedelta(0)

    # A class building tzinfo objects for fixed-offset time zones.
    # Note that FixedOffset(0, "UTC") is a different way to build a
    # UTC tzinfo object.

    class FixedOffset(datetime.tzinfo):
        """Fixed offset in minutes east from UTC."""

        def __init__(self, offset, name=None):
            self.__offset = datetime.timedelta(minutes=offset)
            self.__name = name

        def utcoffset(self, dt):
            return self.__offset

        def tzname(self, dt):
            return self.__name

        def dst(self, dt):
            return _ZERO_TIMEDELTA


    utc = FixedOffset(0, 'UTC')


try:
    from weakref import finalize
except ImportError:
    # Backported from Python 3.6
    import itertools
    from weakref import ref

    class finalize:
        """Class for finalization of weakrefable objects

        finalize(obj, func, *args, **kwargs) returns a callable finalizer
        object which will be called when obj is garbage collected. The
        first time the finalizer is called it evaluates func(*arg, **kwargs)
        and returns the result. After this the finalizer is dead, and
        calling it just returns None.

        When the program exits any remaining finalizers for which the
        atexit attribute is true will be run in reverse order of creation.
        By default atexit is true.
        """

        # Finalizer objects don't have any state of their own.  They are
        # just used as keys to lookup _Info objects in the registry.  This
        # ensures that they cannot be part of a ref-cycle.

        __slots__ = ()
        _registry = {}
        _shutdown = False
        _index_iter = itertools.count()
        _dirty = False
        _registered_with_atexit = False

        class _Info:
            __slots__ = ("weakref", "func", "args", "kwargs", "atexit", "index")

        def __init__(self, obj, func, *args, **kwargs):
            if not self._registered_with_atexit:
                # We may register the exit function more than once because
                # of a thread race, but that is harmless
                import atexit
                atexit.register(self._exitfunc)
                finalize._registered_with_atexit = True
            info = self._Info()
            info.weakref = ref(obj, self)
            info.func = func
            info.args = args
            info.kwargs = kwargs or None
            info.atexit = True
            info.index = next(self._index_iter)
            self._registry[self] = info
            finalize._dirty = True

        def __call__(self, _=None):
            """If alive then mark as dead and return func(*args, **kwargs);
            otherwise return None"""
            info = self._registry.pop(self, None)
            if info and not self._shutdown:
                return info.func(*info.args, **(info.kwargs or {}))

        def detach(self):
            """If alive then mark as dead and return (obj, func, args, kwargs);
            otherwise return None"""
            info = self._registry.get(self)
            obj = info and info.weakref()
            if obj is not None and self._registry.pop(self, None):
                return (obj, info.func, info.args, info.kwargs or {})

        def peek(self):
            """If alive then return (obj, func, args, kwargs);
            otherwise return None"""
            info = self._registry.get(self)
            obj = info and info.weakref()
            if obj is not None:
                return (obj, info.func, info.args, info.kwargs or {})

        @property
        def alive(self):
            """Whether finalizer is alive"""
            return self in self._registry

        @property
        def atexit(self):
            """Whether finalizer should be called at exit"""
            info = self._registry.get(self)
            return bool(info) and info.atexit

        @atexit.setter
        def atexit(self, value):
            info = self._registry.get(self)
            if info:
                info.atexit = bool(value)

        def __repr__(self):
            info = self._registry.get(self)
            obj = info and info.weakref()
            if obj is None:
                return '<%s object at %#x; dead>' % (type(self).__name__, id(self))
            else:
                return '<%s object at %#x; for %r at %#x>' % \
                       (type(self).__name__, id(self), type(obj).__name__, id(obj))

        @classmethod
        def _select_for_exit(cls):
            # Return live finalizers marked for exit, oldest first
            L = [(f,i) for (f,i) in cls._registry.items() if i.atexit]
            L.sort(key=lambda item:item[1].index)
            return [f for (f,i) in L]

        @classmethod
        def _exitfunc(cls):
            # At shutdown invoke finalizers for which atexit is true.
            # This is called once all other non-daemonic threads have been
            # joined.
            reenable_gc = False
            try:
                if cls._registry:
                    import gc
                    if gc.isenabled():
                        reenable_gc = True
                        gc.disable()
                    pending = None
                    while True:
                        if pending is None or finalize._dirty:
                            pending = cls._select_for_exit()
                            finalize._dirty = False
                        if not pending:
                            break
                        f = pending.pop()
                        try:
                            # gc is disabled, so (assuming no daemonic
                            # threads) the following is the only line in
                            # this function which might trigger creation
                            # of a new finalizer
                            f()
                        except Exception:
                            sys.excepthook(*sys.exc_info())
                        assert f not in cls._registry
            finally:
                # prevent any more finalizers from executing during shutdown
                finalize._shutdown = True
                if reenable_gc:
                    gc.enable()


__all__ = ['sys', 'builtins', 'logging.config', 'OrderedDict', 'dictconfig', 'suppress',
           'reduce', 'reload_module', 'Queue', 'PriorityQueue', 'Empty', 'ElementTree', 'ElementTreeParseError',
           'urlretrieve', 'pickle', 'urlencode', 'urlparse', 'unquote', 'quote', 'quote_plus', 'parse_qsl',
           'Enum', 'ConfigParser', 'decimal', 'Decimal', 'DECIMAL_TYPES', 'FixedOffset', 'utc', 'finalize',
           'functools32', 'zip_longest', 'OrderedDict3', 'BrokenPipeError', 'TimeoutError', 'ConnectionResetError',
           'izip', 'accumulate']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..datasource import tensor as astensor
from .core import TensorReduction


class TensorNanProd(operands.NanProd, TensorReduction):
    def __init__(self, axis=None, dtype=None, keepdims=None, combine_size=None, **kw):
        super(TensorNanProd, self).__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims,
                                            _combine_size=combine_size, **kw)

    @staticmethod
    def _get_op_types():
        return TensorNanProd, TensorNanProd, None


def nanprod(a, axis=None, dtype=None, out=None, keepdims=None, combine_size=None):
    """
    Return the product of array elements over a given axis treating Not a
    Numbers (NaNs) as ones.

    One is returned for slices that are all-NaN or empty.

    Parameters
    ----------
    a : array_like
        Tensor containing numbers whose product is desired. If `a` is not an
        tensor, a conversion is attempted.
    axis : int, optional
        Axis along which the product is computed. The default is to compute
        the product of the flattened tensor.
    dtype : data-type, optional
        The type of the returned tensor and of the accumulator in which the
        elements are summed.  By default, the dtype of `a` is used.  An
        exception is when `a` has an integer type with less precision than
        the platform (u)intp. In that case, the default will be either
        (u)int32 or (u)int64 depending on whether the platform is 32 or 64
        bits. For inexact inputs, dtype must be inexact.
    out : Tensor, optional
        Alternate output tensor in which to place the result.  The default
        is ``None``. If provided, it must have the same shape as the
        expected output, but the type will be cast if necessary.  See
        `doc.ufuncs` for details. The casting of NaN to integer can yield
        unexpected results.
    keepdims : bool, optional
        If True, the axes which are reduced are left in the result as
        dimensions with size one. With this option, the result will
        broadcast correctly against the original `arr`.
    combine_size: int, optional
        The number of chunks to combine.

    Returns
    -------
    nanprod : Tensor
        A new tensor holding the result is returned unless `out` is
        specified, in which case it is returned.

    See Also
    --------
    mt.prod : Product across array propagating NaNs.
    isnan : Show which elements are NaN.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.nanprod(1).execute()
    1
    >>> mt.nanprod([1]).execute()
    1
    >>> mt.nanprod([1, mt.nan]).execute()
    1.0
    >>> a = mt.array([[1, 2], [3, mt.nan]])
    >>> mt.nanprod(a).execute()
    6.0
    >>> mt.nanprod(a, axis=0).execute()
    array([ 3.,  2.])

    """
    a = astensor(a)
    if dtype is None:
        dtype = np.nanprod(np.empty((1,), dtype=a.dtype)).dtype
    op = TensorNanProd(axis=axis, dtype=dtype, keepdims=keepdims, combine_size=combine_size)
    return op(a, out=out)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .flip import flip


def flipud(m):
    """
    Flip tensor in the up/down direction.

    Flip the entries in each column in the up/down direction.
    Rows are preserved, but appear in a different order than before.

    Parameters
    ----------
    m : array_like
        Input tensor.

    Returns
    -------
    out : array_like
        A view of `m` with the rows reversed.  Since a view is
        returned, this operation is :math:`\\mathcal O(1)`.

    See Also
    --------
    fliplr : Flip tensor in the left/right direction.
    rot90 : Rotate tensor counterclockwise.

    Notes
    -----
    Equivalent to ``m[::-1,...]``.
    Does not require the tensor to be two-dimensional.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> A = mt.diag([1.0, 2, 3])
    >>> A.execute()
    array([[ 1.,  0.,  0.],
           [ 0.,  2.,  0.],
           [ 0.,  0.,  3.]])
    >>> mt.flipud(A).execute()
    array([[ 0.,  0.,  3.],
           [ 0.,  2.,  0.],
           [ 1.,  0.,  0.]])

    >>> A = mt.random.randn(2,3,5)
    >>> mt.all(mt.flipud(A) == A[::-1,...]).execute()
    True

    >>> mt.flipud([1,2]).execute()
    array([2, 1])

    """
    return flip(m, 0)



from datetime import datetime, timedelta
from gevent.event import Event

from .compat import urlparse, TimeoutError


def _normalize_path(path):
    while '//' in path:
        path = path.replace('//', '/')
    return path.rstrip('/') or '/'


class PathResult(object):
    def __init__(self, key, value=None, dir=False, children=None):
        self.key = key
        self.value = value
        self.dir = dir
        self.children = children or []

    def __repr__(self):
        arg_strs = ['key=%r' % self.key]
        if self.value:
            arg_strs.append('value=%r' % self.value)
        if self.dir:
            arg_strs.append('dir=%r' % self.dir)
        if self.children:
            arg_strs.append('children=%r' % self.children)
        return 'PathResult(' + ', '.join(arg_strs) + ')'


class LocalKVStore(object):
    def __init__(self):
        self._store = dict()
        self._expire_time = dict()
        self._children = {'/': set()}
        self._watch_event = {'/': set()}
        self._watch_event_r = {'/': set()}

    def _read_value(self, item):
        item = _normalize_path(item)
        if item in self._store:
            if item in self._expire_time and self._expire_time[item] < datetime.now():
                del self._store[item]
                raise KeyError(item)
            return PathResult(item, value=self._store[item])
        elif item in self._children:
            return PathResult(item, dir=True)
        else:
            raise KeyError(item)

    def _read_dir(self, item, recursive=False, sort=False):
        item = _normalize_path(item)
        r = PathResult(item, dir=True)
        if recursive:
            for ch in self._children[item]:
                v = self.read(item + '/' + ch, recursive=True)
                if v.dir and v.children:
                    r.children.extend(v.children)
                else:
                    r.children.append(v)
        else:
            r.children = [self._read_value(item + '/' + ch) for ch in self._children[item]]
        if sort:
            r.children = sorted(r.children, key=lambda ch: ch.key)
        return r

    def read(self, item, recursive=False, sort=False):
        item = _normalize_path(item)
        if item in self._store:
            if item in self._expire_time and self._expire_time[item] < datetime.now():
                del self._store[item]
                raise KeyError(item)
            return PathResult(item, value=self._store[item])
        else:
            return self._read_dir(item, recursive=recursive, sort=sort)

    def watch(self, key, timeout=None, recursive=None, sort=False):
        key = _normalize_path(key)
        segments = key.lstrip('/').split('/')
        path = ''
        parent = '/'
        watch_event = Event()
        event_paths = []

        for s in segments:
            if parent not in self._watch_event_r:
                self._watch_event_r[parent] = set()
            if parent not in self._watch_event:
                self._watch_event[parent] = set()
            if recursive:
                event_paths.append(parent)
                self._watch_event_r[parent].add(watch_event)

            path += '/' + s
            parent = path

        if path not in self._watch_event_r:
            self._watch_event_r[path] = set()
        if path not in self._watch_event:
            self._watch_event[path] = set()

        if recursive:
            event_paths.append(key)
            self._watch_event_r[key].add(watch_event)
        else:
            self._watch_event[key].add(watch_event)

        if not watch_event.wait(timeout):
            raise TimeoutError

        if recursive:
            for p in event_paths:
                self._watch_event_r[p].remove(watch_event)
        else:
            self._watch_event[key].remove(watch_event)

        return self.read(key, recursive=recursive, sort=sort)

    def eternal_watch(self, key, recursive=False):
        while True:
            response = self.watch(key, timeout=0, recursive=recursive)
            yield response

    def get_lock(self, lock_name):
        from gevent.lock import RLock
        lock_name = _normalize_path(lock_name)
        if lock_name not in self._store:
            lock = RLock()
            self.write(lock_name, lock)
        else:
            lock = self._store[lock_name]
        return lock

    def write(self, key, value=None, ttl=None, dir=False):
        key = _normalize_path(key)
        segments = key.lstrip('/').split('/')
        path = ''
        parent = '/'
        event_paths = []

        for s in segments:
            if parent not in self._children:
                self._children[parent] = set()
            self._children[parent].add(s)
            event_paths.append(parent)

            path += '/' + s
            if path != key and path in self._store:
                raise KeyError('Not a directory: %s' % key)
            parent = path
        if dir:
            self._children[key] = set()
        else:
            self._store[key] = value
            if ttl:
                self._expire_time[key] = datetime.now() + timedelta(seconds=ttl)

        event_paths.append(key)
        for p in event_paths:
            if p in self._watch_event_r:
                [e.set() for e in list(self._watch_event_r[p])]
        if key in self._watch_event:
            [e.set() for e in list(self._watch_event[key])]

    def delete(self, key, dir=False, recursive=False):
        key = _normalize_path(key)
        if not dir:
            del self._store[key]
        else:
            if not recursive and self._children[key]:
                raise KeyError('Dir %s not empty', key)
            for ch in list(self._children[key]):
                ch_key = key + '/' + ch
                if ch_key in self._store:
                    self.delete(ch_key)
                else:
                    self.delete(ch_key, dir=True, recursive=True)
            del self._children[key]
        dir_name, file_name = key.rsplit('/', 1)
        self._children[dir_name].remove(file_name)

        if key in self._watch_event:
            [e.set() for e in list(self._watch_event[key])]
            del self._watch_event[key]

        if key in self._watch_event_r:
            base_path = ''
            event_paths = ['/']
            for sp in key.lstrip('/').split('/'):
                base_path += '/' + sp
                event_paths.append(base_path)

            for evt in list(self._watch_event_r[key]):
                evt.set()
                for p in event_paths:
                    if evt in self._watch_event_r.get(p, ()):
                        self._watch_event_r[p].remove(evt)
            del self._watch_event_r[key]


class EtcdKVStore(object):
    def __init__(self, etcd_client, base_path=''):
        self._etcd_client = etcd_client
        self._base_path = _normalize_path(base_path)

    def read(self, item, recursive=False):
        from etcd_gevent import EtcdKeyError, EtcdKeyNotFound
        item = _normalize_path(self._base_path + item)

        try:
            val = self._etcd_client.read(item, recursive=recursive)
        except (EtcdKeyError, EtcdKeyNotFound):
            val = None
        if val is None:
            raise KeyError(item)

        r = PathResult(item, value=val.value, dir=val.dir)
        for ch in val.children:
            if ch.key == item:
                continue
            r.children.append(PathResult(ch.key, value=ch.value, dir=ch.dir))
        return r

    def watch(self, item, timeout=None, recursive=None):
        from etcd_gevent import EtcdKeyError, EtcdKeyNotFound, EtcdWatchTimedOut
        item = _normalize_path(self._base_path + item)
        try:
            val = self._etcd_client.watch(item, timeout=timeout, recursive=recursive)
        except (EtcdKeyError, EtcdKeyNotFound):
            raise KeyError(item)
        except EtcdWatchTimedOut:
            raise TimeoutError

        r = PathResult(item, value=val.value, dir=val.dir)
        for ch in val.children:
            r.children.append(PathResult(ch.key, value=ch.value, dir=ch.dir))
        return r

    def eternal_watch(self, key, recursive=False):
        while True:
            response = self.watch(key, timeout=0, recursive=recursive)
            yield response

    def get_lock(self, lock_name):
        from etcd_gevent import Lock
        lock_name = _normalize_path(lock_name)
        return Lock(self._etcd_client, lock_name)

    def write(self, key, value=None, ttl=None, dir=False):
        from etcd_gevent import EtcdNotDir
        key = _normalize_path(self._base_path + key)
        try:
            self._etcd_client.write(key, value, ttl=ttl, dir=dir)
        except EtcdNotDir:
            raise KeyError('%s not dir' % key)

    def delete(self, key, dir=False, recursive=False):
        from etcd_gevent import EtcdNotDir, EtcdNotFile, EtcdDirNotEmpty
        key = _normalize_path(self._base_path + key)
        try:
            self._etcd_client.delete(key, dir=dir, recursive=recursive)
        except (EtcdNotFile, EtcdNotDir):
            raise KeyError(key)
        except EtcdDirNotEmpty:
            raise KeyError('Dir %s not empty', key)


def get(addr):
    if addr == ':inproc:':
        return LocalKVStore()
    parsed = urlparse(addr)
    if parsed.scheme == 'etcd':
        from etcd_gevent.client import Client as EtcdClient
        client = EtcdClient(parsed.hostname, parsed.port)
        return EtcdKVStore(client, parsed.path)
    else:
        raise ValueError('Scheme %s not supported.' % parsed.scheme)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorOutBinOp


class TensorFrexp(operands.Frexp, TensorOutBinOp):
    def __init__(self, casting='same_kind', dtype=None, sparse=False, **kw):
        super(TensorFrexp, self).__init__(_casting=casting,
                                          _dtype=dtype, _sparse=sparse, **kw)

    @property
    def _fun(self):
        return np.frexp


def frexp(x, out1=None, out2=None, out=None, where=None, **kwargs):
    """
    Decompose the elements of x into mantissa and twos exponent.

    Returns (`mantissa`, `exponent`), where `x = mantissa * 2**exponent``.
    The mantissa is lies in the open interval(-1, 1), while the twos
    exponent is a signed integer.

    Parameters
    ----------
    x : array_like
        Tensor of numbers to be decomposed.
    out1 : Tensor, optional
        Output tensor for the mantissa. Must have the same shape as `x`.
    out2 : Tensor, optional
        Output tensor for the exponent. Must have the same shape as `x`.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    (mantissa, exponent) : tuple of tensors, (float, int)
        `mantissa` is a float array with values between -1 and 1.
        `exponent` is an int array which represents the exponent of 2.

    See Also
    --------
    ldexp : Compute ``y = x1 * 2**x2``, the inverse of `frexp`.

    Notes
    -----
    Complex dtypes are not supported, they will raise a TypeError.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> from mars.session import new_session

    >>> x = mt.arange(9)
    >>> y1, y2 = mt.frexp(x)

    >>> sess = new_session().as_default()
    >>> y1_result, y2_result = sess.run(y1, y2)
    >>> y1_result
    array([ 0.   ,  0.5  ,  0.5  ,  0.75 ,  0.5  ,  0.625,  0.75 ,  0.875,
            0.5  ])
    >>> y2_result
    array([0, 1, 2, 2, 3, 3, 3, 3, 4])
    >>> (y1 * 2**y2).execute(session=sess)
    array([ 0.,  1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.])
    """
    op = TensorFrexp(**kwargs)
    return op(x, out1=out1, out2=out2, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np

from .... import opcodes as OperandDef
from ....serialize import StringField
from ....operands import DataSource
from ....compat import izip
from ....config import options
from ....tiles import NotSupportTile
from ..utils import normalize_shape, decide_chunks
from ..core import TensorOperandMixin


class TensorDataSource(DataSource, TensorOperandMixin):
    """
    Tensor data source base class, provide universal tile logic,
    subclass can overwrite tile method.
    """

    __slots__ = ()

    def to_chunk_op(self, *args):
        chunk_shape, idx, chunk_size = args
        chunk_op = self.copy().reset_key()
        chunk_op.params = {'size': chunk_shape, 'index': idx}  # to make op key different
        return chunk_op

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]

        chunk_size = tensor.params.raw_chunks or options.tensor.chunks
        chunk_size = decide_chunks(tensor.shape, chunk_size, tensor.dtype.itemsize)
        chunk_size_idxes = (range(len(size)) for size in chunk_size)

        out_chunks = []
        for chunk_shape, chunk_idx in izip(itertools.product(*chunk_size),
                                           itertools.product(*chunk_size_idxes)):
            chunk_op = op.to_chunk_op(chunk_shape, chunk_idx, chunk_size)
            out_chunk = chunk_op.new_chunk(None, chunk_shape, index=chunk_idx)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, tensor.shape, chunks=out_chunks, nsplits=chunk_size)


class TensorNoInput(TensorDataSource):
    """
    Tensor operand with no inputs.
    """

    def check_inputs(self, inputs):
        # no inputs
        if inputs and len(inputs) > 0:
            raise ValueError("Tensor data source has no inputs")

    def __call__(self, shape, chunks=None):
        shape = normalize_shape(shape)
        return self.new_tensor(None, shape, raw_chunks=chunks)


class TensorHasInput(TensorDataSource):
    """
    Tensor operand with a single input.
    """

    @property
    def input(self):
        return self._input

    def check_inputs(self, inputs):
        # no inputs
        if len(inputs) != 1:
            raise ValueError("Tensor can only have 1 input")

    def _set_inputs(self, inputs):
        super(TensorHasInput, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    @classmethod
    def tile(cls, op):
        out_chunks = []
        for c in op.input.chunks:
            out_chunk = op.copy().reset_key().new_chunk([c], c.shape, index=c.index)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, op.outputs[0].shape, chunks=out_chunks,
                                  nsplits=op.input.nsplits)

    def __call__(self, a):
        return self.new_tensor([a], a.shape)


class TensorLike(TensorHasInput):
    def _set_inputs(self, inputs):
        super(TensorLike, self)._set_inputs(inputs)
        if self.dtype is None:
            self._dtype = self.input.dtype
        if self.gpu is None:
            self._gpu = self.input.op.gpu

        # FIXME: remove when cupy supports other dtypes
        if self._gpu and self._dtype not in (np.float32, np.float64):
            raise NotImplementedError('Sparse tensor on GPU only supports float32 and float64')


class TensorFetchChunk(TensorNoInput, TensorOperandMixin):
    _op_type_ = OperandDef.FETCH

    _to_fetch_key = StringField('to_fetch_key')

    def __init__(self, dtype=None, to_fetch_key=None, **kw):
        super(TensorFetchChunk, self).__init__(_dtype=dtype, _to_fetch_key=to_fetch_key,
                                               **kw)

    @classmethod
    def tile(cls, op):
        raise NotSupportTile('FetchChunk is a chunk operand which does not support tile')



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import array

from .. import compat
from ..compat import six
from ..cluster_info import HasClusterInfoActor
from ..utils import classproperty
from ..promise import PromiseActor


class OperandState(compat.Enum):
    UNSCHEDULED = 'unscheduled'
    READY = 'ready'
    RUNNING = 'running'
    FINISHED = 'finished'
    CACHED = 'cached'
    FREED = 'freed'
    FATAL = 'fatal'
    CANCELLING = 'cancelling'
    CANCELLED = 'cancelled'

    @classproperty
    def STORED_STATES(self):
        """
        States on which the data of the operand is stored
        """
        return self.FINISHED, self.CACHED

    @classproperty
    def SUCCESSFUL_STATES(self):
        """
        States on which the operand is executed successfully
        """
        return self.FINISHED, self.CACHED, self.FREED

    @classproperty
    def TERMINATED_STATES(self):
        """
        States on which the operand has already terminated
        """
        return self.FINISHED, self.CACHED, self.FREED, self.FATAL, self.CANCELLED


class GraphState(compat.Enum):
    UNSCHEDULED = 'unscheduled'
    PREPARING = 'preparing'
    RUNNING = 'running'
    SUCCEEDED = 'succeeded'
    FAILED = 'failed'
    CANCELLING = 'cancelling'
    CANCELLED = 'cancelled'

    @classproperty
    def TERMINATED_STATES(self):
        """
        States on which the graph has already terminated
        """
        return self.SUCCEEDED, self.FAILED


class SchedulerActor(HasClusterInfoActor, PromiseActor):
    def __init__(self, **kwargs):
        super(SchedulerActor, self).__init__()

        if 'balancer' in kwargs:
            self._balancer_ref = kwargs['balancer']

    @classmethod
    def default_name(cls):
        return 's:{0}'.format(cls.__name__)


def remove_shuffle_chunks(chunks):
    return list(n for n in chunks if not getattr(n.op, '_shuffle_source', False))


if six.PY3:
    def array_to_bytes(typecode, initializer):
        return array.array(typecode, initializer).tobytes()
else:
    def array_to_bytes(typecode, initializer):
        return array.array(typecode, initializer).tostring()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .... import operands
from ....tiles import NotSupportTile
from ..core import TensorOperandMixin


class TensorFuseChunk(operands.Fuse, TensorOperandMixin):
    def __init__(self, dtype=None, **kw):
        super(TensorFuseChunk, self).__init__(_dtype=dtype, **kw)

    @classmethod
    def tile(cls, op):
        raise NotSupportTile('FetchChunk is a chunk operand which does not support tile')


class TensorFuseChunkMixin(TensorOperandMixin):
    __slots__ = ()

    @classmethod
    def tile(cls, op):
        raise NotSupportTile('FetchChunk is a chunk operand which does not support tile')

    def __call__(self, fuse_chunks):
        head_chunk = fuse_chunks[0]
        tail_chunk = fuse_chunks[-1]
        setattr(self, '_operands', [c.op for c in fuse_chunks])
        return self.new_chunk(head_chunk.inputs, tail_chunk.shape,
                              _composed=fuse_chunks, _key=tail_chunk.key)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import infer_dtype
from .core import TensorBinOp, TensorConstant


class TensorPower(operands.Power, TensorBinOp):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorPower, self).__init__(_casting=casting, _err=err,
                                          _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        return x1.issparse() and not x2.issparse()

    @classmethod
    def constant_cls(cls):
        return TensorPowConstant


class TensorPowConstant(operands.PowConstant, TensorConstant):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorPowConstant, self).__init__(_casting=casting, _err=err,
                                                _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        if hasattr(x1, 'issparse') and x1.issparse():
            return True
        return False


@infer_dtype(np.power)
def power(x1, x2, out=None, where=None, **kwargs):
    """
    First tensor elements raised to powers from second tensor, element-wise.

    Raise each base in `x1` to the positionally-corresponding power in
    `x2`.  `x1` and `x2` must be broadcastable to the same shape. Note that an
    integer type raised to a negative integer power will raise a ValueError.

    Parameters
    ----------
    x1 : array_like
        The bases.
    x2 : array_like
        The exponents.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : Tensor
        The bases in `x1` raised to the exponents in `x2`.

    See Also
    --------
    float_power : power function that promotes integers to float

    Examples
    --------
    Cube each element in a list.

    >>> import mars.tensor as mt

    >>> x1 = range(6)
    >>> x1
    [0, 1, 2, 3, 4, 5]
    >>> mt.power(x1, 3).execute()
    array([  0,   1,   8,  27,  64, 125])

    Raise the bases to different exponents.

    >>> x2 = [1.0, 2.0, 3.0, 3.0, 2.0, 1.0]
    >>> mt.power(x1, x2).execute()
    array([  0.,   1.,   8.,  27.,  16.,   5.])

    The effect of broadcasting.

    >>> x2 = mt.array([[1, 2, 3, 3, 2, 1], [1, 2, 3, 3, 2, 1]])
    >>> x2.execute()
    array([[1, 2, 3, 3, 2, 1],
           [1, 2, 3, 3, 2, 1]])
    >>> mt.power(x1, x2).execute()
    array([[ 0,  1,  8, 27, 16,  5],
           [ 0,  1,  8, 27, 16,  5]])
    """
    op = TensorPower(**kwargs)
    return op(x1, x2, out=out, where=where)


@infer_dtype(np.power, reverse=True)
def rpower(x1, x2, **kwargs):
    op = TensorPower(**kwargs)
    return op.rcall(x1, x2)



# Copyright 2009 Brian Quinlan. All Rights Reserved.
# Licensed to PSF under a Contributor Agreement.

"""Execute computations asynchronously using threads or processes."""

__author__ = 'Brian Quinlan (brian@sweetapp.com)'

from ._base import (FIRST_COMPLETED,
                                      FIRST_EXCEPTION,
                                      ALL_COMPLETED,
                                      CancelledError,
                                      TimeoutError,
                                      Future,
                                      Executor,
                                      wait,
                                      as_completed)
from .thread import ThreadPoolExecutor

try:
    from .process import ProcessPoolExecutor
except ImportError:
    # some platforms don't have multiprocessing
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ..datasource import tensor as astensor


def ravel(a):
    """Return a contiguous flattened array.

    A 1-D array, containing the elements of the input, is returned.  A copy is
    made only if needed.

    Parameters
    ----------
    a : array_like
        Input tensor.  The elements in `a` are packed as a 1-D tensor.

    Returns
    -------
    y : array_like
        If `a` is a matrix, y is a 1-D tensor, otherwise y is a tensor of
        the same subtype as `a`. The shape of the returned array is
        ``(a.size,)``. Matrices are special cased for backward
        compatibility.

    See Also
    --------
    Tensor.flat : 1-D iterator over an array.
    Tensor.flatten : 1-D array copy of the elements of an array
                      in row-major order.
    Tensor.reshape : Change the shape of an array without changing its data.

    Examples
    --------
    It is equivalent to ``reshape(-1)``.

    >>> import mars.tensor as mt

    >>> x = mt.array([[1, 2, 3], [4, 5, 6]])
    >>> print(mt.ravel(x).execute())
    [1 2 3 4 5 6]

    >>> print(x.reshape(-1).execute())
    [1 2 3 4 5 6]

    >>> print(mt.ravel(x.T).execute())
    [1 4 2 5 3 6]

    >>> a = mt.arange(12).reshape(2,3,2).swapaxes(1,2); a.execute()
    array([[[ 0,  2,  4],
            [ 1,  3,  5]],
           [[ 6,  8, 10],
            [ 7,  9, 11]]])
    >>> a.ravel().execute()
    array([ 0,  2,  4,  1,  3,  5,  6,  8, 10,  7,  9, 11])

    """
    a = astensor(a)
    if a.ndim == 0:
        return a[np.newaxis]
    return a.reshape(-1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ....operands import fft as fftop
from ....compat import izip
from ..datasource import tensor as astensor
from .core import TensorFFTNMixin, validate_fftn


class TensorFFTN(fftop.FFTN, TensorFFTNMixin):
    def __init__(self, shape=None, axes=None, norm=None, dtype=None, **kw):
        super(TensorFFTN, self).__init__(_shape=shape, _axes=axes, _norm=norm,
                                         _dtype=dtype, **kw)

    @classmethod
    def _get_shape(cls, op, shape):
        new_shape = list(shape)
        if op.shape is not None:
            for ss, axis in izip(op.shape, op.axes):
                new_shape[axis] = ss
        return tuple(new_shape)

    def _set_inputs(self, inputs):
        super(TensorFFTN, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, a):
        shape = self._get_shape(self, a.shape)
        return self.new_tensor([a], shape)


def fftn(a, s=None, axes=None, norm=None):
    """
    Compute the N-dimensional discrete Fourier Transform.

    This function computes the *N*-dimensional discrete Fourier Transform over
    any number of axes in an *M*-dimensional tensor by means of the Fast Fourier
    Transform (FFT).

    Parameters
    ----------
    a : array_like
        Input tensor, can be complex.
    s : sequence of ints, optional
        Shape (length of each transformed axis) of the output
        (``s[0]`` refers to axis 0, ``s[1]`` to axis 1, etc.).
        This corresponds to ``n`` for ``fft(x, n)``.
        Along any axis, if the given shape is smaller than that of the input,
        the input is cropped.  If it is larger, the input is padded with zeros.
        if `s` is not given, the shape of the input along the axes specified
        by `axes` is used.
    axes : sequence of ints, optional
        Axes over which to compute the FFT.  If not given, the last ``len(s)``
        axes are used, or all axes if `s` is also not specified.
        Repeated indices in `axes` means that the transform over that axis is
        performed multiple times.
    norm : {None, "ortho"}, optional
        Normalization mode (see `mt.fft`). Default is None.

    Returns
    -------
    out : complex Tensor
        The truncated or zero-padded input, transformed along the axes
        indicated by `axes`, or by a combination of `s` and `a`,
        as explained in the parameters section above.

    Raises
    ------
    ValueError
        If `s` and `axes` have different length.
    IndexError
        If an element of `axes` is larger than than the number of axes of `a`.

    See Also
    --------
    mt.fft : Overall view of discrete Fourier transforms, with definitions
        and conventions used.
    ifftn : The inverse of `fftn`, the inverse *n*-dimensional FFT.
    fft : The one-dimensional FFT, with definitions and conventions used.
    rfftn : The *n*-dimensional FFT of real input.
    fft2 : The two-dimensional FFT.
    fftshift : Shifts zero-frequency terms to centre of tensor

    Notes
    -----
    The output, analogously to `fft`, contains the term for zero frequency in
    the low-order corner of all axes, the positive frequency terms in the
    first half of all axes, the term for the Nyquist frequency in the middle
    of all axes and the negative frequency terms in the second half of all
    axes, in order of decreasingly negative frequency.

    See `mt.fft` for details, definitions and conventions used.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.mgrid[:3, :3, :3][0]
    >>> mt.fft.fftn(a, axes=(1, 2)).execute()
    array([[[  0.+0.j,   0.+0.j,   0.+0.j],
            [  0.+0.j,   0.+0.j,   0.+0.j],
            [  0.+0.j,   0.+0.j,   0.+0.j]],
           [[  9.+0.j,   0.+0.j,   0.+0.j],
            [  0.+0.j,   0.+0.j,   0.+0.j],
            [  0.+0.j,   0.+0.j,   0.+0.j]],
           [[ 18.+0.j,   0.+0.j,   0.+0.j],
            [  0.+0.j,   0.+0.j,   0.+0.j],
            [  0.+0.j,   0.+0.j,   0.+0.j]]])
    >>> mt.fft.fftn(a, (2, 2), axes=(0, 1)).execute()
    array([[[ 2.+0.j,  2.+0.j,  2.+0.j],
            [ 0.+0.j,  0.+0.j,  0.+0.j]],
           [[-2.+0.j, -2.+0.j, -2.+0.j],
            [ 0.+0.j,  0.+0.j,  0.+0.j]]])

    >>> import matplotlib.pyplot as plt
    >>> [X, Y] = mt.meshgrid(2 * mt.pi * mt.arange(200) / 12,
    ...                      2 * mt.pi * mt.arange(200) / 34)
    >>> S = mt.sin(X) + mt.cos(Y) + mt.random.uniform(0, 1, X.shape)
    >>> FS = mt.fft.fftn(S)
    >>> plt.imshow(mt.log(mt.abs(mt.fft.fftshift(FS))**2).execute())
    <matplotlib.image.AxesImage object at 0x...>
    >>> plt.show()

    """
    a = astensor(a)
    axes = validate_fftn(a, s=s, axes=axes, norm=norm)
    op = TensorFFTN(shape=s, axes=axes, norm=norm, dtype=np.dtype(np.complex_))
    return op(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from collections import Iterable

import numpy as np

from .... import operands
from ....compat import izip
from ..utils import unify_chunks
from ..core import TensorOperandMixin
from ..arithmetic.utils import tree_add
from ..datasource import tensor as astensor


class TensorTensorDot(operands.TensorDot, TensorOperandMixin):
    def __init__(self, a_axes=None, b_axes=None, dtype=None, sparse=False, **kw):
        super(TensorTensorDot, self).__init__(_a_axes=a_axes, _b_axes=b_axes,
                                              _dtype=dtype, _sparse=sparse, **kw)

    def _set_inputs(self, inputs):
        super(TensorTensorDot, self)._set_inputs(inputs)
        self._a = self._inputs[0]
        self._b = self._inputs[1]

    def __call__(self, a, b):
        shape = tuple(s for i, s in enumerate(a.shape) if i not in set(self._a_axes)) + \
            tuple(s for i, s in enumerate(b.shape) if i not in set(self._b_axes))
        return self.new_tensor([a, b], shape)

    @classmethod
    def tile(cls, op):
        a, b, a_axes, b_axes = op.a, op.b, op.a_axes, op.b_axes

        c = itertools.count(max(a.ndim, b.ndim))
        a_ax = tuple(a_axes.index(i) if i in a_axes else next(c) for i in range(a.ndim))
        b_ax = tuple(b_axes.index(i) if i in b_axes else next(c) for i in range(b.ndim))
        a, b = unify_chunks((a, a_ax), (b, b_ax))

        a_output_indexes = [range(len(a.nsplits[i])) for i in range(a.ndim) if i not in a_axes]
        b_output_indexes = [range(len(b.nsplits[i])) for i in range(b.ndim) if i not in b_axes]
        output_axes = [(0, i) for i in range(a.ndim) if i not in a_axes] + \
                      [(1, i) for i in range(b.ndim) if i not in b_axes]

        out_chunks = []
        for out_idx in itertools.product(*itertools.chain(a_output_indexes, b_output_indexes)):
            a_indexes = [None] * a.ndim
            b_indexes = [None] * b.ndim
            tensor_shape = []
            for i, idx in enumerate(out_idx):
                t_idx, axis = output_axes[i]
                t = (a, b)[t_idx]
                (a_indexes if t_idx == 0 else b_indexes)[axis] = idx
                tensor_shape.append(t.nsplits[axis][idx])
            tensor_shape = tuple(tensor_shape)

            tensordot_chunks = []
            for contract_indexes in itertools.product(*[range(len(a.nsplits[ax])) for ax in a_axes]):
                a_indices, b_indices = list(a_indexes), list(b_indexes)
                for a_axis, contract_index in izip(a_axes, contract_indexes):
                    a_indices[a_axis] = contract_index
                for b_axis, contract_index in izip(b_axes, contract_indexes):
                    b_indices[b_axis] = contract_index

                tensordot_chunk_op = op.copy().reset_key()
                tensordot_chunk = tensordot_chunk_op.new_chunk(
                    [a.cix[tuple(a_indices)], b.cix[tuple(b_indices)]], tensor_shape)
                tensordot_chunks.append(tensordot_chunk)

            if len(tensordot_chunks) == 1:
                c = tensordot_chunks[0]
                chunk_op = c.op.copy()
                chunk = chunk_op.new_chunk(c.inputs, c.shape, index=out_idx)
            else:
                chunk = tree_add(op.dtype, tensordot_chunks, out_idx, tensor_shape)
            out_chunks.append(chunk)

        get_nsplits = lambda t_idx, i: (a, b)[t_idx].nsplits[i]
        nsplits = [get_nsplits(*it) for it in output_axes]
        new_op = op.copy()
        return new_op.new_tensors([a, b], op.outputs[0].shape,
                                  chunks=out_chunks, nsplits=nsplits)


def tensordot(a, b, axes=2, sparse=None):
    """
    Compute tensor dot product along specified axes for tensors >= 1-D.
    Given two tensors (arrays of dimension greater than or equal to one),
    `a` and `b`, and an array_like object containing two array_like
    objects, ``(a_axes, b_axes)``, sum the products of `a`'s and `b`'s
    elements (components) over the axes specified by ``a_axes`` and
    ``b_axes``. The third argument can be a single non-negative
    integer_like scalar, ``N``; if it is such, then the last ``N``
    dimensions of `a` and the first ``N`` dimensions of `b` are summed
    over.
    Parameters
    ----------
    a, b : array_like, len(shape) >= 1
        Tensors to "dot".
    axes : int or (2,) array_like
        * integer_like
          If an int N, sum over the last N axes of `a` and the first N axes
          of `b` in order. The sizes of the corresponding axes must match.
        * (2,) array_like
          Or, a list of axes to be summed over, first sequence applying to `a`,
          second to `b`. Both elements array_like must be of the same length.
    See Also
    --------
    dot, einsum
    Notes
    -----
    Three common use cases are:
        * ``axes = 0`` : tensor product :math:`a\\otimes b`
        * ``axes = 1`` : tensor dot product :math:`a\\cdot b`
        * ``axes = 2`` : (default) tensor double contraction :math:`a:b`
    When `axes` is integer_like, the sequence for evaluation will be: first
    the -Nth axis in `a` and 0th axis in `b`, and the -1th axis in `a` and
    Nth axis in `b` last.
    When there is more than one axis to sum over - and they are not the last
    (first) axes of `a` (`b`) - the argument `axes` should consist of
    two sequences of the same length, with the first axis to sum over given
    first in both sequences, the second axis second, and so forth.
    Examples
    --------
    >>> import mars.tensor as mt

    A "traditional" example:
    >>> a = mt.arange(60.).reshape(3,4,5)
    >>> b = mt.arange(24.).reshape(4,3,2)
    >>> c = mt.tensordot(a,b, axes=([1,0],[0,1]))
    >>> c.shape
    (5, 2)
    >>> r = c.execute()
    >>> r
    array([[ 4400.,  4730.],
           [ 4532.,  4874.],
           [ 4664.,  5018.],
           [ 4796.,  5162.],
           [ 4928.,  5306.]])
    >>> # A slower but equivalent way of computing the same...
    >>> ra = np.arange(60.).reshape(3,4,5)
    >>> rb = np.arange(24.).reshape(4,3,2)
    >>> d = np.zeros((5,2))
    >>> for i in range(5):
    ...   for j in range(2):
    ...     for k in range(3):
    ...       for n in range(4):
    ...         d[i,j] += ra[k,n,i] * rb[n,k,j]
    >>> r == d
    array([[ True,  True],
           [ True,  True],
           [ True,  True],
           [ True,  True],
           [ True,  True]], dtype=bool)
    An extended example taking advantage of the overloading of + and \\*:
    >>> a = mt.array(range(1, 9))
    >>> a.shape = (2, 2, 2)
    >>> A = mt.array(('a', 'b', 'c', 'd'), dtype=object)
    >>> A.shape = (2, 2)
    >>> a.execute(); A.execute()
    array([[[1, 2],
            [3, 4]],
           [[5, 6],
            [7, 8]]])
    array([[a, b],
           [c, d]], dtype=object)
    >>> mt.tensordot(a, A).execute() # third argument default is 2 for double-contraction
    array([abbcccdddd, aaaaabbbbbbcccccccdddddddd], dtype=object)
    >>> mt.tensordot(a, A, 1).execute()
    array([[[acc, bdd],
            [aaacccc, bbbdddd]],
           [[aaaaacccccc, bbbbbdddddd],
            [aaaaaaacccccccc, bbbbbbbdddddddd]]], dtype=object)
    >>> mt.tensordot(a, A, 0).execute() # tensor product (result too long to incl.)
    array([[[[[a, b],
              [c, d]],
              ...
    >>> mt.tensordot(a, A, (0, 1)).execute()
    array([[[abbbbb, cddddd],
            [aabbbbbb, ccdddddd]],
           [[aaabbbbbbb, cccddddddd],
            [aaaabbbbbbbb, ccccdddddddd]]], dtype=object)
    >>> mt.tensordot(a, A, (2, 1)).execute()
    array([[[abb, cdd],
            [aaabbbb, cccdddd]],
           [[aaaaabbbbbb, cccccdddddd],
            [aaaaaaabbbbbbbb, cccccccdddddddd]]], dtype=object)
    >>> mt.tensordot(a, A, ((0, 1), (0, 1))).execute()
    array([abbbcccccddddddd, aabbbbccccccdddddddd], dtype=object)
    >>> mt.tensordot(a, A, ((2, 1), (1, 0))).execute()
    array([acccbbdddd, aaaaacccccccbbbbbbdddddddd], dtype=object)
    """
    a = astensor(a)
    b = astensor(b)

    if isinstance(axes, Iterable):
        a_axes, b_axes = axes
    else:
        a_axes = tuple(range(a.ndim - 1, a.ndim - axes - 1, -1))
        b_axes = tuple(range(0, axes))

    if isinstance(a_axes, Iterable):
        a_axes = tuple(a_axes)
    else:
        a_axes = (a_axes,)
    a_axes = tuple(axis if axis >= 0 else a.ndim + axis for axis in a_axes)
    if isinstance(b_axes, Iterable):
        b_axes = tuple(b_axes)
    else:
        b_axes = (b_axes,)
    b_axes = tuple(axis if axis >= 0 else b.ndim + axis for axis in b_axes)

    if a.shape and b.shape and \
            not np.array_equal(np.array(a.shape)[list(a_axes)], np.array(b.shape)[list(b_axes)]):
        raise ValueError('shape-mismatch for sum')

    sparse = sparse if sparse is not None else a.issparse() and b.issparse()
    op = TensorTensorDot(a_axes=a_axes, b_axes=b_axes, dtype=np.promote_types(a.dtype, b.dtype),
                         sparse=sparse)
    return op(a, b)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import Iterable

import numpy as np

from ....operands import UnravelIndex
from ...core import ExecutableTuple
from ..core import TensorOperandMixin
from ..datasource import tensor as astensor


class TensorUnravelIndex(UnravelIndex, TensorOperandMixin):
    def __init__(self, dims=None, dtype=None, **kw):
        super(TensorUnravelIndex, self).__init__(_dims=dims, _dtype=dtype, **kw)

    def _set_inputs(self, inputs):
        super(TensorUnravelIndex, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, indices):
        kws = [{'pos': i} for i in range(len(self._dims))]
        return ExecutableTuple(self.new_tensors([indices], indices.shape, kws=kws, output_limit=len(kws)))

    @classmethod
    def tile(cls, op):
        indices = op.inputs[0]
        dims = op.dims

        out_chunks = [list() for _ in range(len(dims))]
        for in_chunk in indices.chunks:
            chunk_op = op.copy().reset_key()
            chunk_kws = [{'pos': i, 'index': in_chunk.index} for i in range(len(dims))]
            chunks = chunk_op.new_chunks([in_chunk], in_chunk.shape, kws=chunk_kws,
                                         output_limit=len(dims))
            for out_chunk, c in zip(out_chunks, chunks):
                out_chunk.append(c)

        new_op = op.copy()
        shapes = [t.shape for t in op.outputs]
        kws = [{'chunks': out_chunk, 'nsplits': indices.nsplits} for out_chunk in out_chunks]
        return new_op.new_tensors(op.inputs, shapes, kws=kws, output_limit=len(dims))


def unravel_index(indices, dims):
    """
    Converts a flat index or tensor of flat indices into a tuple
    of coordinate tensors.

    Parameters
    ----------
    indices : array_like
        An integer tensor whose elements are indices into the flattened
        version of a tensor of dimensions ``dims``.
    dims : tuple of ints
        The shape of the tensor to use for unraveling ``indices``.

    Returns
    -------
    unraveled_coords : tuple of Tensor
        Each tensor in the tuple has the same shape as the ``indices``
        tensor.

    See Also
    --------
    ravel_multi_index

    Examples
    --------
    >>> import mars.tensor as mt
    >>> from mars.session import new_session

    >>> sess = new_session().as_default()

    >>> sess.run(mt.unravel_index([22, 41, 37], (7,6)))
    (array([3, 6, 6]), array([4, 5, 1]))

    >>> sess.run(mt.unravel_index(1621, (6,7,8,9)))
    (3, 1, 4, 1)
    """
    indices = astensor(indices)
    if isinstance(dims, Iterable):
        dims = tuple(dims)
    else:
        dims = (dims,)

    op = TensorUnravelIndex(dims=dims, dtype=np.dtype(np.intp))
    return op(indices)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ....operands import fft as fftop
from ....tiles import NotSupportTile
from ..core import TensorOperandMixin
from ..datasource import arange
from .core import TensorFFTMixin


class TensorFFTFreq(fftop.FFTFreq, TensorFFTMixin):
    def __init__(self, n=None, d=None, dtype=None, gpu=False, **kw):
        super(TensorFFTFreq, self).__init__(_n=n, _d=d, _dtype=dtype, _gpu=gpu, **kw)

    def __call__(self, chunks=None):
        shape = (self.n,)
        return self.new_tensor(None, shape, raw_chunks=chunks)

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]
        in_tensor = arange(op.n, gpu=op.gpu, dtype=op.dtype,
                           chunks=tensor.params.raw_chunks).single_tiles()

        out_chunks = []
        for c in in_tensor.chunks:
            chunk_op = TensorFFTFreqChunk(n=op.n, d=op.d, dtype=op.dtype)
            out_chunk = chunk_op.new_chunk([c], c.shape, index=c.index)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, tensor.shape,
                                  chunks=out_chunks, nsplits=in_tensor.nsplits,
                                  **tensor.params)


class TensorFFTFreqChunk(fftop.FFTFreqChunk, TensorOperandMixin):
    def __init__(self, n=None, d=None, dtype=None, **kw):
        super(TensorFFTFreqChunk, self).__init__(_n=n, _d=d, _dtype=dtype, **kw)

    def _set_inputs(self, inputs):
        super(TensorFFTFreqChunk, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    @classmethod
    def tile(cls, op):
        raise NotSupportTile('FFTFreqChunk is a chunk operand which does not support tile')


def fftfreq(n, d=1.0, gpu=False, chunks=None):
    """
    Return the Discrete Fourier Transform sample frequencies.

    The returned float tensor `f` contains the frequency bin centers in cycles
    per unit of the sample spacing (with zero at the start).  For instance, if
    the sample spacing is in seconds, then the frequency unit is cycles/second.

    Given a window length `n` and a sample spacing `d`::

      f = [0, 1, ...,   n/2-1,     -n/2, ..., -1] / (d*n)   if n is even
      f = [0, 1, ..., (n-1)/2, -(n-1)/2, ..., -1] / (d*n)   if n is odd

    Parameters
    ----------
    n : int
        Window length.
    d : scalar, optional
        Sample spacing (inverse of the sampling rate). Defaults to 1.
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    chunks : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension

    Returns
    -------
    f : Tensor
        Array of length `n` containing the sample frequencies.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> signal = mt.array([-2, 8, 6, 4, 1, 0, 3, 5], dtype=float)
    >>> fourier = mt.fft.fft(signal)
    >>> n = signal.size
    >>> timestep = 0.1
    >>> freq = mt.fft.fftfreq(n, d=timestep)
    >>> freq.execute()
    array([ 0.  ,  1.25,  2.5 ,  3.75, -5.  , -3.75, -2.5 , -1.25])

    """
    n, d = int(n), float(d)
    op = TensorFFTFreq(n=n, d=d, dtype=np.dtype(float), gpu=gpu)
    return op(chunks)
